"""Core recipes for DFTB+"""
from __future__ import annotations

from typing import TYPE_CHECKING

from quacc import job
from quacc.recipes.dftb._base import base_fn

if TYPE_CHECKING:
    from typing import Literal

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import RunSchema


@job
def static_job(
    atoms: Atoms,
    method: Literal["GFN1-xTB", "GFN2-xTB", "DFTB"] = "GFN2-xTB",
    kpts: tuple | list[tuple] | dict | None = None,
    **calc_kwargs,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    method
        Method to use.
    kpts
        k-point grid to use.
    **calc_kwargs
        Custom kwargs for the calculator that would override the
        calculator defaults. Set a value to `None` to remove a pre-existing key
        entirely. For a list of available keys, refer to the
        `ase.calculators.dftb.Dftb` calculator.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {
        "Hamiltonian_": "xTB" if "xtb" in method.lower() else "DFTB",
        "Hamiltonian_MaxSccIterations": 200,
        "Hamiltonian_Method": method if "xtb" in method.lower() else None,
        "kpts": kpts or ((1, 1, 1) if atoms.pbc.any() else None),
    }

    return base_fn(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "DFTB+ Static"},
    )


@job
def relax_job(
    atoms: Atoms,
    method: Literal["GFN1-xTB", "GFN2-xTB", "DFTB"] = "GFN2-xTB",
    kpts: tuple | list[tuple] | dict | None = None,
    relax_cell: bool = False,
    **calc_kwargs,
) -> RunSchema:
    """
    Carry out a structure relaxation.

    Parameters
    ----------
    atoms
        Atoms object
    method
        Method to use.
    kpts
        k-point grid to use.
    relax_cell
        Whether to relax the unit cell shape/volume in addition to the
        positions.
    **calc_kwargs
        Custom kwargs for the calculator that would override the
        calculator defaults. Set a value to `None` to remove a pre-existing key
        entirely. For a list of available keys, refer to the
        `ase.calculators.dftb.Dftb` calculator.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {
        "Driver_": "GeometryOptimization",
        "Driver_AppendGeometries": "Yes",
        "Driver_LatticeOpt": "Yes" if relax_cell else "No",
        "Driver_MaxSteps": 2000,
        "Hamiltonian_": "xTB" if "xtb" in method.lower() else "DFTB",
        "Hamiltonian_MaxSccIterations": 200,
        "Hamiltonian_Method": method if "xtb" in method.lower() else None,
        "kpts": kpts or ((1, 1, 1) if atoms.pbc.any() else None),
    }

    return base_fn(
        atoms,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "DFTB+ Relax"},
    )



"""Slab recipes for EMT."""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from quacc import flow
from quacc.atoms.slabs import make_slabs_from_bulk
from quacc.recipes.common.slabs import bulk_to_slabs_subflow
from quacc.recipes.emt.core import relax_job, static_job

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import OptSchema, RunSchema


@flow
def bulk_to_slabs_flow(
    atoms: Atoms,
    make_slabs_kwargs: dict[str, Any] | None = None,
    run_static: bool = True,
    slab_relax_kwargs: dict[str, Any] | None = None,
    slab_static_kwargs: dict[str, Any] | None = None,
) -> list[RunSchema | OptSchema]:
    """
    Workflow consisting of:

    1. Slab generation

    2. Slab relaxations

    3. Slab statics (optional)

    Parameters
    ----------
    atoms
        Atoms object
    make_slabs_kwargs
        Additional keyword arguments to pass to
        [quacc.atoms.slabs.make_slabs_from_bulk][]
    run_static
        Whether to run the static calculation.
    slab_relax_kwargs
        Additional keyword arguments to pass to [quacc.recipes.emt.core.relax_job][].
    slab_static_kwargs
        Additional keyword arguments to pass to [quacc.recipes.emt.core.static_job][].

    Returns
    -------
    list[RunSchema | OptSchema]
        [RunSchema][quacc.schemas.ase.summarize_run] or
        [OptSchema][quacc.schemas.ase.summarize_opt_run] for each slab.
    """

    make_slabs_kwargs = make_slabs_kwargs or {}
    slab_relax_kwargs = slab_relax_kwargs or {}
    slab_static_kwargs = slab_static_kwargs or {}

    return bulk_to_slabs_subflow(
        atoms,
        partial(relax_job, **slab_relax_kwargs),
        static_job=partial(static_job, **slab_static_kwargs) if run_static else None,
        make_slabs_fn=partial(make_slabs_from_bulk, **make_slabs_kwargs),
    )



"""Utility functions for dealing with Atoms."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np

from quacc.atoms.core import copy_atoms, get_atoms_id

if TYPE_CHECKING:
    from ase.atoms import Atoms

logger = logging.getLogger(__name__)


def prep_next_run(
    atoms: Atoms, assign_id: bool = True, move_magmoms: bool = True
) -> Atoms:
    """
    Prepares the Atoms object for a new run.

    Depending on the arguments, this function will:
        - Move the converged magnetic moments to the initial magnetic moments.
        - Assign a unique ID to the Atoms object in atoms.info["_id"]. Any
          existing IDs will be moved to atoms.info["_old_ids"].

    In all cases, the calculator will be reset so new jobs can be run.

    Parameters
    ----------
    atoms
        Atoms object
    assign_id
        Whether to assign a unique ID to the Atoms object in atoms.info["_id"].
        Any existing IDs will be moved to atoms.info["_old_ids"].
    move_magmoms
        If True, move atoms.calc.results["magmoms"] to
        atoms.get_initial_magnetic_moments()

    Returns
    -------
    Atoms
        Updated Atoms object.
    """
    atoms = copy_atoms(atoms)

    if (
        move_magmoms
        and hasattr(atoms, "calc")
        and getattr(atoms.calc, "results", None) is not None
    ):
        # If there are initial magmoms set, then we should see what the final
        # magmoms are. If they are present, move them to initial. If they are
        # not present, it means the calculator doesn't support the "magmoms"
        # property so we have to retain the initial magmoms given no further
        # info.
        if atoms.has("initial_magmoms"):
            atoms.set_initial_magnetic_moments(
                atoms.calc.results.get("magmoms", atoms.get_initial_magnetic_moments())
            )
        # If there are no initial magmoms set, just check the results and set
        # everything to 0.0 if there is nothing there.
        else:
            atoms.set_initial_magnetic_moments(
                atoms.calc.results.get("magmoms", [0.0] * len(atoms))
            )

    # Clear off the calculator so we can run a new job. If we don't do this,
    # then something like atoms *= (2,2,2) still has a calculator attached,
    # which is a bit confusing.
    atoms.calc = None

    # Give the Atoms object a unique ID. This will be helpful for querying
    # later. Also store any old IDs somewhere else for future reference. Note:
    # Keep this at the end of the function so that the ID is assigned based on
    # the returned Atoms object.
    if assign_id:
        if atoms.info.get("_id", None) is not None:
            if atoms.info.get("_old_ids") is None:
                atoms.info["_old_ids"] = []
            atoms.info["_old_ids"].append(atoms.info["_id"])
        atoms.info["_id"] = get_atoms_id(atoms)

    return atoms


def set_magmoms(
    atoms: Atoms,
    elemental_mags_dict: dict[str, float] | None = None,
    elemental_mags_default: float = 1.0,
    copy_magmoms: bool = True,
    mag_cutoff: float | None = 0.05,
) -> Atoms:  # sourcery skip
    """
    Sets the initial magnetic moments in the Atoms object.

    This function deserves particular attention. The following logic is applied:
    - If there is a converged set of magnetic moments, those are moved to the
    initial magmoms if copy_magmoms is True. - If there is no converged set of
    magnetic moments but the user has set initial magmoms, those are simply used
    as is. - If there are no converged magnetic moments or initial magnetic
    moments, then the default magnetic moments from the preset
    elemental_mags_dict (if specified) are set as the initial magnetic moments.
    - For any of the above scenarios, if mag_cutoff is not None, the newly set
    initial magnetic moments are checked. If all have a magnitude below
    mag_cutoff, then they are all set to 0 (no spin polarization).

    Parameters
    ----------
    atoms
        Atoms object
    elemental_mags_dict
        Dictionary of elemental symbols and their corresponding magnetic moments
        to set. If None, no default values will be used.
    elemental_mags_default
        Default magnetic moment on an element if no magnetic moment is specified
        in the elemental_mags_dict. Only used if elemental_mags_dict is not
        None. This kwarg is mainly a convenience so that you don't need to list
        every single element in the elemental_mags_dict.
    copy_magmoms
        Whether to copy the magnetic moments from the converged set of magnetic
        moments to the initial magnetic moments.
    mag_cutoff
        Magnitude below which the magnetic moments are considered to be zero. If
        None, no cutoff will be applied

    Returns
    -------
    Atoms
        Atoms object
    """

    # Handle the magnetic moments Check if a prior job was run and pull the
    # prior magmoms
    if hasattr(atoms, "calc") and getattr(atoms.calc, "results", None) is not None:
        mags = atoms.calc.results.get("magmoms", [0.0] * len(atoms))
        # Note: It is important that we set mags to 0.0 here rather than None if
        # the calculator has no magmoms because: 1) ispin=1 might be set, and 2)
        # we do not want the preset magmoms to be used.
    else:
        mags = None

    # Check if the user has set any initial magmoms
    has_initial_mags = atoms.has("initial_magmoms")

    # If there are no initial magmoms set and this is not a follow-up job, we
    # may need to add some from the preset yaml.
    if mags is None:
        if not has_initial_mags:
            # If the preset dictionary has default magmoms, set those by
            # element. If the element isn't in the magmoms dict then set it to
            # mag_default.
            if elemental_mags_dict:
                initial_mags = np.array(
                    [
                        elemental_mags_dict.get(atom.symbol, elemental_mags_default)
                        for atom in atoms
                    ]
                )
                atoms.set_initial_magnetic_moments(initial_mags)
        else:
            pass
    elif copy_magmoms:
        atoms.set_initial_magnetic_moments(mags)

    # If all the set mags are below mag_cutoff, set them to 0
    if mag_cutoff:
        has_new_initial_mags = atoms.has("initial_magmoms")
        new_initial_mags = atoms.get_initial_magnetic_moments()
        if has_new_initial_mags and np.all(np.abs(new_initial_mags) < mag_cutoff):
            atoms.set_initial_magnetic_moments([0.0] * len(atoms))

    return atoms



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms


def parse_pp_and_cutoff(config: dict[str, Any], atoms: Atoms) -> dict[str, Any] | None:
    """
    Function that parses the pseudopotentials and cutoffs from a preset file.
    The cutoffs are taken from the largest value of the cutoffs among the elements
    present.

    Parameters
    ----------
    config
        The config dictionary return by load_yaml_calc

    atoms
        The atoms object to get the elements from

    Returns
    -------
    dict | None
        A dictionary containing the pseudopotentials and cutoffs
    """

    if "pseudopotentials" not in config:
        return None

    pp_dict = config["pseudopotentials"]
    unique_elements = list(set(atoms.symbols))
    wfc_cutoff, rho_cutoff = 0, 0
    pseudopotentials = {}
    for element in unique_elements:
        if pp_dict[element]["cutoff_wfc"] > wfc_cutoff:
            wfc_cutoff = pp_dict[element]["cutoff_wfc"]
        if pp_dict[element]["cutoff_rho"] > rho_cutoff:
            rho_cutoff = pp_dict[element]["cutoff_rho"]
        pseudopotentials[element] = pp_dict[element]["filename"]
    tmp_input_data = {"system": {"ecutwfc": wfc_cutoff, "ecutrho": rho_cutoff}}

    return {"pseudopotentials": pseudopotentials, "input_data": tmp_input_data}



"""Phonon recipes for TBLite"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from quacc import flow
from quacc.recipes.common.phonons import phonon_flow as phonon_flow_
from quacc.recipes.tblite.core import static_job

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from numpy.typing import ArrayLike

    from quacc.schemas._aliases.phonons import PhononSchema


@flow
def phonon_flow(
    atoms: Atoms,
    supercell_matrix: ArrayLike = ((2, 0, 0), (0, 2, 0), (0, 0, 2)),
    atom_disp: float = 0.01,
    symprec: float = 1e-5,
    t_step: float = 10,
    t_min: float = 0,
    t_max: float = 1000,
    static_job_kwargs: dict[str, Any] | None = None,
) -> PhononSchema:
    """
    Carry out a phonon calculation.

    Parameters
    ----------
    atoms
        Atoms object
    supercell_matrix
        Supercell matrix to use. Defaults to 2x2x2 supercell.
    atom_disp
        Atomic displacement (A).
    symprec
        Precision for symmetry detection.
    t_step
        Temperature step (K).
    t_min
        Min temperature (K).
    t_max
        Max temperature (K).
    static_job_kwargs
        Keyword arguments for [quacc.recipes.tblite.core.static_job][]
        for the force calculations.

    Returns
    -------
    PhononSchema
        Dictionary of results from [quacc.schemas.phonons.summarize_phonopy][]
    """
    static_job_kwargs = static_job_kwargs or {}

    return phonon_flow_(
        atoms,
        partial(static_job, **static_job_kwargs),
        supercell_matrix=supercell_matrix,
        atom_disp=atom_disp,
        t_step=t_step,
        t_min=t_min,
        t_max=t_max,
        phonopy_kwargs={"symprec": symprec},
        additional_fields={"name": "TBLite Phonons"},
    )



"""Schemas for storing metadata about Atoms objects."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from ase.atoms import Atom, Atoms
from emmet.core.structure import MoleculeMetadata, StructureMetadata
from monty.json import jsanitize
from pymatgen.io.ase import AseAtomsAdaptor

from quacc.atoms.core import copy_atoms
from quacc.utils.dicts import remove_dict_nones, sort_dict

if TYPE_CHECKING:
    from typing import Any

    from quacc.schemas._aliases.atoms import AtomsSchema


def atoms_to_metadata(
    atoms: Atoms,
    charge_and_multiplicity: tuple[int, int] | None = None,
    get_metadata: bool = True,
    store_pmg: bool = True,
    additional_fields: dict[str, Any] | None = None,
) -> AtomsSchema:
    """
    Convert an ASE Atoms object to a dict suitable for storage in MongoDB.

    Parameters
    ----------
    atoms
        ASE Atoms object to store in {"atoms": atoms}
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    get_metadata
        Whether to store atoms metadata in the returned dict.
    store_pmg
        Whether to store the Pymatgen Structure/Molecule object in {"structure":
        Structure} or {"molecule": Molecule}, respectively.
    additional_fields
        Additional fields to add to the document.

    Returns
    -------
    AtomsSchema
        Dict of metadata about the Atoms object.
    """

    additional_fields = additional_fields or {}
    atoms = copy_atoms(atoms)
    results = {}

    # Get any charge or multiplicity keys
    if charge_and_multiplicity:
        atoms.charge = charge_and_multiplicity[0]
        atoms.spin_multiplicity = charge_and_multiplicity[1]

    # Strip the dummy atoms, if present
    del atoms[[atom.index for atom in atoms if atom.symbol == "X"]]

    # Get Atoms metadata, if requested. emmet already has built-in tools for
    # generating pymatgen Structure/Molecule metadata, so we'll just use that.
    if get_metadata:
        if atoms.pbc.any():
            struct = AseAtomsAdaptor().get_structure(atoms)
            metadata = StructureMetadata().from_structure(struct).model_dump()
            if store_pmg:
                results["structure"] = struct
        else:
            mol = AseAtomsAdaptor().get_molecule(atoms, charge_spin_check=False)
            metadata = MoleculeMetadata().from_molecule(mol).model_dump()
            if store_pmg:
                results["molecule"] = mol
    else:
        metadata = {}

    # Copy the info flags as a separate entry in the DB for easy querying
    results["atoms_info"] = _quacc_sanitize(atoms.info)

    # Store Atoms object
    results["atoms"] = atoms

    # Combine the metadata and results dictionaries
    atoms_doc_unsorted = metadata | results | additional_fields

    return sort_dict(remove_dict_nones(atoms_doc_unsorted))


def _quacc_sanitize(obj: Any) -> Any:
    """
    Sanitizes an object for storage in MongoDB.

    This is an analogue of monty's jsanitize function but meant to serialize
    Atom/Atoms objects as well.

    Parameters
    ----------
    obj
        Object to sanitize

    Returns
    -------
    Any
        Sanitized object
    """
    if isinstance(obj, (Atom, Atoms)):
        obj = atoms_to_metadata(obj)
    elif isinstance(obj, (list, tuple, np.ndarray)):
        obj = [_quacc_sanitize(i) for i in obj]
    elif isinstance(obj, dict):
        obj = {k.__str__(): _quacc_sanitize(v) for k, v in obj.items()}
    else:
        obj = jsanitize(obj)
    return obj



"""Core recipes for VASP."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from quacc import job
from quacc.recipes.vasp._base import base_fn

if TYPE_CHECKING:
    from typing import Any

    from ase import Atoms

    from quacc.schemas._aliases.vasp import DoubleRelaxSchema, VaspSchema


@job
def static_job(
    atoms: Atoms,
    preset: str | None = "BulkSet",
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    preset
        Preset to use from `quacc.calculators.vasp.presets`.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `quacc.calculators.vasp.vasp.Vasp` calculator.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    calc_defaults = {
        "ismear": -5,
        "laechg": True,
        "lcharg": True,
        "lreal": False,
        "lwave": True,
        "nedos": 5001,
        "nsw": 0,
    }
    return base_fn(
        atoms,
        preset=preset,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "VASP Static"},
        copy_files=copy_files,
    )


@job
def relax_job(
    atoms: Atoms,
    preset: str | None = "BulkSet",
    relax_cell: bool = True,
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Relax a structure.

    Parameters
    ----------
    atoms
        Atoms object
    preset
        Preset to use from `quacc.calculators.vasp.presets`.
    relax_cell
        True if a volume relaxation (ISIF = 3) should be performed. False if
        only the positions (ISIF = 2) should be updated.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `quacc.calculators.vasp.vasp.Vasp` calculator.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    calc_defaults = {
        "ediffg": -0.02,
        "isif": 3 if relax_cell else 2,
        "ibrion": 2,
        "isym": 0,
        "lcharg": False,
        "lwave": False,
        "nsw": 200,
        "symprec": 1e-8,
    }
    return base_fn(
        atoms,
        preset=preset,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "VASP Relax"},
        copy_files=copy_files,
    )


@job
def double_relax_job(
    atoms: Atoms,
    preset: str | None = "BulkSet",
    relax_cell: bool = True,
    relax1_kwargs: dict[str, Any] | None = None,
    relax2_kwargs: dict[str, Any] | None = None,
) -> DoubleRelaxSchema:
    """
    Double-relax a structure. This is particularly useful for a few reasons:

    1. To carry out a cheaper pre-relaxation before the high-quality run.

    2. To carry out a GGA calculation before a meta-GGA or hybrid calculation
    that requires the GGA wavefunction.

    3. To carry out volume relaxations where large changes in volume
    can require a second relaxation to resolve forces.

    Parameters
    ----------
    atoms
        Atoms object
    preset
        Preset to use from `quacc.calculators.vasp.presets`.
    relax_cell
        True if a volume relaxation (ISIF = 3) should be performed. False if
        only the positions (ISIF = 2) should be updated.
    relax1_kwargs
        Dictionary of custom kwargs for the first relaxation.
    relax2_kwargs
        Dictionary of custom kwargs for the second relaxation.

    Returns
    -------
    DoubleRelaxSchema
        Dictionary of results
    """
    relax1_kwargs = relax1_kwargs or {}
    relax2_kwargs = relax2_kwargs or {}

    # Run first relaxation
    summary1 = relax_job.__wrapped__(
        atoms, preset=preset, relax_cell=relax_cell, **relax1_kwargs
    )

    # Run second relaxation
    summary2 = relax_job.__wrapped__(
        summary1["atoms"],
        preset=preset,
        relax_cell=relax_cell,
        copy_files=[Path(summary1["dir_name"]) / "WAVECAR"],
        **relax2_kwargs,
    )
    summary2["relax1"] = summary1

    return summary2



from __future__ import annotations

from ase.io.espresso import Namelist

PH_KEYS = Namelist(
    {
        "INPUTPH": [
            "amass",
            "outdir",
            "prefix",
            "niter_ph",
            "tr2_ph",
            "alpha_mix",
            "nmix_ph",
            "verbosity",
            "reduce_io",
            "max_seconds",
            "dftd3_hess",
            "fildyn",
            "fildrho",
            "fildvscf",
            "epsil",
            "lrpa",
            "lnoloc",
            "trans",
            "lraman",
            "eth_rps",
            "eth_ns",
            "dek",
            "recover",
            "low_directory_check",
            "only_init",
            "qplot",
            "q2d",
            "q_in_band_form",
            "electron_phonon",
            "el_ph_nsigma",
            "el_ph_sigma",
            "ahc_dir",
            "ahc_nbnd",
            "ahc_nbndskip",
            "skip_upperfan",
            "lshift_q",
            "zeu",
            "zue",
            "elop",
            "fpol",
            "ldisp",
            "nogg",
            "asr",
            "ldiag",
            "lqdir",
            "search_sym",
            "nq1",
            "nq2",
            "nq3",
            "nk1",
            "nk2",
            "nk3",
            "k1",
            "k2",
            "k3",
            "diagonalization",
            "read_dns_bare",
            "ldvscf_interpolate",
            "wpot_dir",
            "do_long_range",
            "do_charge_neutral",
            "start_irr",
            "last_irr",
            "nat_todo",
            "modenum",
            "start_q",
            "last_q",
            "dvscf_star",
            "drho_star",
        ]
    }
)

NEB_KEYS = Namelist(
    {
        "PATH": [
            "string_method",
            "restart_mode",
            "nstep_path",
            "num_of_images",
            "opt_scheme",
            "CI_scheme",
            "first_last_opt",
            "minimum_image",
            "temp_req",
            "ds",
            "k_max",
            "k_min",
            "path_thr",
            "uses_masses",
            "use_freezing",
            "lfcp",
            "fcp_mu",
            "fcp_thr",
            "fcp_scheme",
        ]
    }
)

Q2R_KEYS = Namelist({"INPUT": ["fildyn", "flfrc", "zasr", "loto_2d", "write_lr"]})

MATDYN_KEYS = Namelist(
    {
        "INPUT": [
            "flfrc",
            "asr",
            "huang",
            "dos",
            "nk1",
            "nk2",
            "nk3",
            "deltaE",
            "ndos",
            "degauss",
            "fldos",
            "flfrq",
            "flvec",
            "fleig",
            "fldyn",
            "at",
            "l1",
            "l2",
            "l3",
            "ntyp",
            "amass",
            "readtau",
            "fltau",
            "la2F",
            "q_in_band_form",
            "q_in_cryst_coord",
            "eigen_similarity",
            "fd",
            "na_ifc",
            "nosym",
            "loto_2d",
            "loto_disable",
            "read_lr",
            "write_frc",
        ]
    }
)

DYNMAT_KEYS = Namelist(
    {
        "INPUT": [
            "fildyn",
            "q",
            "amass",
            "asr",
            "remove_interaction_blocks",
            "axis",
            "lperm",
            "lplasma",
            "filout",
            "fileig",
            "filmol",
            "filxsf",
            "loto_2d",
            "el_ph_nsig",
            "el_ph_sigma",
        ]
    }
)

PROJWFC_KEYS = Namelist(
    {
        "PROJWFC": [
            "prefix",
            "outdir",
            "ngauss",
            "degauss",
            "emin",
            "emax",
            "deltae",
            "lsymdiag_basis",
            "pawproj",
            "filpdos",
            "filproj",
            "lwrite_overlaps",
            "lbinary_data",
            "kresolveddos",
            "tdosinboxes",
            "n_proj_boxes",
            "irmin",
            "irmax",
            "plot_boxes",
        ]
    }
)


# Updated pw.x keys from my own branch

PW_KEYS = Namelist(
    (
        (
            "CONTROL",
            [
                "calculation",
                "title",
                "verbosity",
                "restart_mode",
                "wf_collect",
                "nstep",
                "iprint",
                "tstress",
                "tprnfor",
                "dt",
                "outdir",
                "wfcdir",
                "prefix",
                "lkpoint_dir",
                "max_seconds",
                "etot_conv_thr",
                "forc_conv_thr",
                "disk_io",
                "pseudo_dir",
                "tefield",
                "dipfield",
                "lelfield",
                "nberrycyc",
                "lorbm",
                "lberry",
                "gdir",
                "nppstr",
                "gatelfcp",
                "trism",
            ],
        ),
        (
            "SYSTEM",
            [
                "ibrav",
                "celldm",
                "a",
                "b",
                "c",
                "cosAB",
                "cosAC",
                "cosBCnat",
                "ntyp",
                "nbnd",
                "tot_charge",
                "tot_magnetization",
                "starting_magnetization",
                "ecutwfc",
                "ecutrho",
                "ecutfock",
                "nr1",
                "nr2",
                "nr3",
                "nr1s",
                "nr2s",
                "nr3s",
                "nosym",
                "nosym_evc",
                "noinv",
                "no_t_rev",
                "force_symmorphic",
                "use_all_frac",
                "occupations",
                "one_atom_occupations",
                "starting_spin_angle",
                "degauss",
                "smearing",
                "nspin",
                "noncolin",
                "ecfixed",
                "qcutz",
                "q2sigma",
                "input_dft",
                "exx_fraction",
                "screening_parameter",
                "exxdiv_treatment",
                "x_gamma_extrapolation",
                "ecutvcut",
                "nqx1",
                "nqx2",
                "nqx3",
                "lda_plus_u",
                "lda_plus_u_kind",
                "Hubbard_U",
                "Hubbard_J0",
                "Hubbard_occ",
                "Hubbard_alpha",
                "Hubbard_beta",
                "Hubbard_J",
                "starting_ns_eigenvalue",
                "U_projection_type",
                "edir",
                "emaxpos",
                "eopreg",
                "eamp",
                "angle1",
                "angle2",
                "constrained_magnetization",
                "fixed_magnetization",
                "lambda",
                "report",
                "lspinorb",
                "assume_isolated",
                "esm_bc",
                "esm_w",
                "esm_efield",
                "esm_nfit",
                "fcp_mu",
                "vdw_corr",
                "london",
                "london_s6",
                "london_c6",
                "london_rvdw",
                "london_rcut",
                "ts_vdw_econv_thr",
                "ts_vdw_isolated",
                "xdm",
                "xdm_a1",
                "xdm_a2",
                "space_group",
                "uniqueb",
                "origin_choice",
                "rhombohedral",
                "zmon",
                "realxz",
                "block",
                "block_1",
                "block_2",
                "block_height",
            ],
        ),
        (
            "ELECTRONS",
            [
                "electron_maxstep",
                "scf_must_converge",
                "conv_thr",
                "adaptive_thr",
                "conv_thr_init",
                "conv_thr_multi",
                "mixing_mode",
                "mixing_beta",
                "mixing_ndim",
                "mixing_fixed_ns",
                "diagonalization",
                "ortho_para",
                "diago_thr_init",
                "diago_cg_maxiter",
                "diago_david_ndim",
                "diago_full_acc",
                "efield",
                "efield_cart",
                "efield_phase",
                "startingpot",
                "startingwfc",
                "tqr",
            ],
        ),
        (
            "IONS",
            [
                "ion_dynamics",
                "ion_positions",
                "pot_extrapolation",
                "wfc_extrapolation",
                "remove_rigid_rot",
                "ion_temperature",
                "tempw",
                "tolp",
                "delta_t",
                "nraise",
                "refold_pos",
                "upscale",
                "bfgs_ndim",
                "trust_radius_max",
                "trust_radius_min",
                "trust_radius_ini",
                "w_1",
                "w_2",
            ],
        ),
        (
            "CELL",
            [
                "cell_dynamics",
                "press",
                "wmass",
                "cell_factor",
                "press_conv_thr",
                "cell_dofree",
            ],
        ),
        (
            "RISM",
            [
                "nsolv",
                "closure",
                "tempv",
                "ecutsolv",
                "solute_lj",
                "solute_epsilon",
                "solute_sigma",
                "starting1d",
                "starting3d",
                "smear1d",
                "smear3d",
                "rism1d_maxstep",
                "rism3d_maxstep",
                "rism1d_conv_thr",
                "rism3d_conv_thr",
                "mdiis1d_size",
                "mdiis3d_size",
                "mdiis1d_step",
                "mdiis3d_step",
                "rism1d_bond_width",
                "rism1d_dielectric",
                "rism1d_molesize",
                "rism1d_nproc",
                "rism3d_conv_level",
                "rism3d_planar_average",
                "laue_nfit",
                "laue_expand_right",
                "laue_expand_left",
                "laue_starting_right",
                "laue_starting_left",
                "laue_buffer_right",
                "laue_buffer_left",
                "laue_both_hands",
                "laue_wall",
                "laue_wall_z",
                "laue_wall_rho",
                "laue_wall_epsilon",
                "laue_wall_sigma",
                "laue_wall_lj6",
            ],
        ),
    )
)

ALL_KEYS = {
    "pw": PW_KEYS,
    "ph": PH_KEYS,
    "neb": NEB_KEYS,
    "q2r": Q2R_KEYS,
    "dynmat": DYNMAT_KEYS,
    "projwfc": PROJWFC_KEYS,
}



"""Summarizer for phonopy"""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from quacc import SETTINGS
from quacc.schemas.atoms import atoms_to_metadata
from quacc.utils.dicts import merge_several_dicts, sort_dict
from quacc.utils.files import get_uri
from quacc.wflow_tools.db import results_to_db

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from maggma.core import Store
    from phonopy import Phonopy

    from quacc.schemas._aliases.phonons import PhononSchema


def summarize_phonopy(
    phonon: Phonopy,
    input_atoms: Atoms | None = None,
    parameters: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> PhononSchema:
    """
    Summarize a Phonopy object.

    Parameters
    ----------
    phonon
        Phonopy object
    input_atoms
        Input atoms object
    parameters
        Calculator parameters used to generate the phonon object.
    additional_fields
        Additional fields to add to the document.
    store
        Whether to store the document in the database.

    Returns
    -------
    PhononSchema
        The PhononSchema.
    """
    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    uri = get_uri(Path.cwd())

    inputs = {
        "parameters": parameters,
        "phonopy_metadata": {"version": phonon.version},
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }

    results = {
        "results": {
            "thermal_properties": phonon.get_thermal_properties_dict(),
            "mesh_properties": phonon.get_mesh_dict(),
            "force_constants": phonon.force_constants,
        }
    }
    phonon.save(settings={"force_constants": True})

    atoms_metadata = atoms_to_metadata(input_atoms) if input_atoms else {}
    unsorted_task_doc = merge_several_dicts(
        atoms_metadata, inputs, results, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc



"""Recipes for slabs."""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from quacc import flow, job
from quacc.atoms.slabs import make_adsorbate_structures, make_slabs_from_bulk
from quacc.recipes.common.slabs import bulk_to_slabs_subflow, slab_to_ads_subflow
from quacc.recipes.vasp._base import base_fn

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.vasp import VaspSchema


@job
def slab_static_job(
    atoms: Atoms,
    preset: str | None = "SlabSet",
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Function to carry out a single-point calculation on a slab.

    Parameters
    ----------
    atoms
        Atoms object
    preset
        Preset to use from `quacc.calculators.vasp.presets`.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `quacc.calculators.vasp.vasp.Vasp` calculator.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    calc_defaults = {
        "auto_dipole": True,
        "ismear": -5,
        "laechg": True,
        "lcharg": True,
        "lreal": False,
        "lvhar": True,
        "lwave": True,
        "nedos": 5001,
        "nsw": 0,
    }
    return base_fn(
        atoms,
        preset=preset,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "VASP Slab Static"},
        copy_files=copy_files,
    )


@job
def slab_relax_job(
    atoms: Atoms,
    preset: str | None = "SlabSet",
    copy_files: str | Path | list[str | Path] | None = None,
    **calc_kwargs,
) -> VaspSchema:
    """
    Function to relax a slab.

    Parameters
    ----------
    atoms
        Atoms object
    preset
        Preset to use from `quacc.calculators.vasp.presets`.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.
    **calc_kwargs
        Custom kwargs for the Vasp calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `quacc.calculators.vasp.vasp.Vasp` calculator.

    Returns
    -------
    VaspSchema
        Dictionary of results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    calc_defaults = {
        "auto_dipole": True,
        "ediffg": -0.02,
        "isif": 2,
        "ibrion": 2,
        "isym": 0,
        "lcharg": False,
        "lwave": False,
        "nsw": 200,
        "symprec": 1e-8,
    }
    return base_fn(
        atoms,
        preset=preset,
        calc_defaults=calc_defaults,
        calc_swaps=calc_kwargs,
        additional_fields={"name": "VASP Slab Relax"},
        copy_files=copy_files,
    )


@flow
def bulk_to_slabs_flow(
    atoms: Atoms,
    make_slabs_kwargs: dict[str, Any] | None = None,
    run_static: bool = True,
    slab_relax_kwargs: dict[str, Any] | None = None,
    slab_static_kwargs: dict[str, Any] | None = None,
) -> list[VaspSchema]:
    """
    Workflow consisting of:

    1. Slab generation

    2. Slab relaxations

    3. Slab statics (optional)

    Parameters
    ----------
    atoms
        Atoms object
    make_slabs_kwargs
        Additional keyword arguments to pass to [quacc.atoms.slabs.make_slabs_from_bulk][]
    run_static
        Whether to run the static calculation.
    slab_relax_kwargs
        Additional keyword arguments to pass to [quacc.recipes.vasp.slabs.slab_relax_job][].
    slab_static_kwargs
        Additional keyword arguments to pass to [quacc.recipes.vasp.slabs.slab_static_job][].

    Returns
    -------
    list[VaspSchema]
        List of dictionary results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    make_slabs_kwargs = make_slabs_kwargs or {}
    slab_relax_kwargs = slab_relax_kwargs or {}
    slab_static_kwargs = slab_static_kwargs or {}

    relax_job = partial(slab_relax_job, **slab_relax_kwargs)
    static_job = partial(slab_static_job, **slab_static_kwargs)
    make_slabs_fn = partial(make_slabs_from_bulk, **make_slabs_kwargs)

    return bulk_to_slabs_subflow(
        atoms,
        relax_job,
        static_job=static_job if run_static else None,
        make_slabs_fn=make_slabs_fn,
    )


@flow
def slab_to_ads_flow(
    slab: Atoms,
    adsorbate: Atoms,
    make_ads_kwargs: dict[str, Any] | None = None,
    run_static: bool = True,
    slab_relax_kwargs: dict[str, Any] | None = None,
    slab_static_kwargs: dict[str, Any] | None = None,
) -> list[VaspSchema]:
    """
    Workflow consisting of:

    1. Slab-adsorbate generation

    2. Slab-adsorbate relaxations

    3. Slab-adsorbate statics (optional)

    Parameters
    ----------
    slab
        Atoms object for the slab structure.
    adsorbate
        Atoms object for the adsorbate.
    make_ads_kwargs
        Additional keyword arguments to pass to [quacc.atoms.slabs.make_adsorbate_structures][]
    run_static
        Whether to run the static calculation.
    slab_relax_kwargs
        Additional keyword arguments to pass to [quacc.recipes.vasp.slabs.slab_relax_job][].
    slab_static_kwargs
        Additional keyword arguments to pass to [quacc.recipes.vasp.slabs.slab_static_job][].

    Returns
    -------
    list[VaspSchema]
        List of dictionaries of results from [quacc.schemas.vasp.vasp_summarize_run][]
    """

    make_ads_kwargs = make_ads_kwargs or {}
    slab_relax_kwargs = slab_relax_kwargs or {}
    slab_static_kwargs = slab_static_kwargs or {}

    return slab_to_ads_subflow(
        slab,
        adsorbate,
        partial(slab_relax_job, **slab_relax_kwargs),
        static_job=partial(slab_static_job, **slab_static_kwargs)
        if run_static
        else None,
        make_ads_fn=partial(make_adsorbate_structures, **make_ads_kwargs),
    )



"""Q-Chem calculator IO."""
from __future__ import annotations

import struct
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ase import units
from emmet.core.tasks import _parse_custodian
from monty.io import zopen
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.io.qchem.inputs import QCInput
from pymatgen.io.qchem.outputs import QCOutput
from pymatgen.io.qchem.sets import QChemDictSet

if TYPE_CHECKING:
    from typing import Any, Literal

    from ase.atoms import Atoms

    from quacc.calculators.qchem.qchem import Results


def write_qchem(
    atoms: Atoms,
    directory: Path | str = ".",
    charge: int = 0,
    spin_multiplicity: int = 1,
    basis_set: str = "def2-tzvpd",
    job_type: Literal["sp", "force", "opt", "freq"] = "force",
    scf_algorithm: str = "diis",
    qchem_input_params: dict[str, Any] | None = None,
    prev_orbital_coeffs: list[float] | None = None,
) -> None:
    """
    Write the Q-Chem input files.

    Parameters
    ----------
    atoms
        The Atoms object to be used for the calculation.
    directory
        The directory in which to write the Q-Chem input files.
    charge
        The total charge of the molecular system.
    spin_multiplicity
        The spin multiplicity of the molecular system.
    basis_set
        The basis set to use for the calculation.
    job_type
        The type of calculation to perform.
    scf_algorithm
        The SCF algorithm to use for the calculation.
    qchem_input_params
        Dictionary of Q-Chem input parameters to be passed to
        `pymatgen.io.qchem.sets.DictSet`.
    prev_orbital_coeffs
        The orbital coefficients from a previous calculation.

    Returns
    -------
    None
    """
    directory = Path(directory)

    if prev_orbital_coeffs is not None:
        with Path(directory / "53.0").open(mode="wb") as file:
            for val in prev_orbital_coeffs:
                data = struct.pack("d", val)
                file.write(data)
        if "overwrite_inputs" not in qchem_input_params:
            qchem_input_params["overwrite_inputs"] = {}
        if "rem" not in qchem_input_params["overwrite_inputs"]:
            qchem_input_params["overwrite_inputs"]["rem"] = {}
        if "scf_guess" not in qchem_input_params["overwrite_inputs"]["rem"]:
            qchem_input_params["overwrite_inputs"]["rem"]["scf_guess"] = "read"

    atoms.charge = charge
    atoms.spin_multiplicity = spin_multiplicity
    mol = AseAtomsAdaptor.get_molecule(atoms)
    QChemDictSet(
        mol, job_type, basis_set, scf_algorithm, qchem_version=6, **qchem_input_params
    ).write(directory / "mol.qin")


def read_qchem(directory: Path | str = ".") -> tuple[Results, list[float]]:
    """
    Read Q-Chem log files.

    Parameters
    ----------
    directory
        The directory in which the Q-Chem calculation was run.

    Returns
    -------
    tuple[Results, list[float]]
        The results of the calculation and the orbital coefficients from a previous
        calculation.
    """
    directory = Path(directory)
    qc_input = QCInput.from_file(directory / "mol.qin").as_dict()
    qc_output = QCOutput(directory / "mol.qout").data

    results: Results = {
        "energy": qc_output["final_energy"] * units.Hartree,
        "qc_output": qc_output,
        "qc_input": qc_input,
        "custodian": _parse_custodian(directory),
    }

    # Read the gradient scratch file in 8 byte chunks
    grad_scratch = directory / "131.0"
    if grad_scratch.exists() and grad_scratch.stat().st_size > 0:
        tmp_grad_data = []
        with zopen(grad_scratch, mode="rb") as file:
            binary = file.read()
        tmp_grad_data.extend(
            struct.unpack("d", binary[ii * 8 : (ii + 1) * 8])[0]
            for ii in range(len(binary) // 8)
        )
        grad = [
            [
                float(tmp_grad_data[ii * 3]),
                float(tmp_grad_data[ii * 3 + 1]),
                float(tmp_grad_data[ii * 3 + 2]),
            ]
            for ii in range(len(tmp_grad_data) // 3)
        ]

        # Ensure that the scratch values match the correct values from the
        # output file but with higher precision
        if qc_output["pcm_gradients"] is not None:
            gradient = qc_output["pcm_gradients"][0]
        else:
            gradient = qc_output["gradients"][0]
        for ii, subgrad in enumerate(grad):
            for jj, val in enumerate(subgrad):
                if abs(gradient[ii, jj] - val) > 1e-6:
                    raise ValueError(
                        "Difference between gradient value in scratch file vs. output file should not be this large."
                    )
                gradient[ii, jj] = val

        # Convert gradient to force + deal with units
        results["forces"] = gradient * (-units.Hartree / units.Bohr)

    # Read Hessian scratch file in 8 byte chunks
    hessian_scratch = directory / "132.0"
    if hessian_scratch.exists() and hessian_scratch.stat().st_size > 0:
        tmp_hess_data = []
        with zopen(hessian_scratch, mode="rb") as file:
            binary = file.read()
        tmp_hess_data.extend(
            struct.unpack("d", binary[ii * 8 : (ii + 1) * 8])[0]
            for ii in range(len(binary) // 8)
        )
        reshaped_hess = np.reshape(
            np.array(tmp_hess_data),
            (len(qc_output["species"]) * 3, len(qc_output["species"]) * 3),
        )
        results["hessian"] = reshaped_hess * (units.Hartree / units.Bohr**2)
        results["enthalpy"] = qc_output["total_enthalpy"] * (units.kcal / units.mol)
        results["entropy"] = qc_output["total_entropy"] * (
            0.001 * units.kcal / units.mol
        )

    # Read orbital coefficients scratch file in 8 byte chunks
    prev_orbital_coeffs = []
    with zopen(directory / "53.0", mode="rb") as file:
        binary = file.read()
    prev_orbital_coeffs.extend(
        struct.unpack("d", binary[ii * 8 : (ii + 1) * 8])[0]
        for ii in range(len(binary) // 8)
    )

    return results, prev_orbital_coeffs



"""Base jobs for ORCA"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.orca import ORCA, OrcaProfile, OrcaTemplate

from quacc import SETTINGS
from quacc.runners.ase import run_calc
from quacc.schemas.cclib import cclib_summarize_run
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.cclib import cclibSchema

_LABEL = OrcaTemplate()._label  # skipcq: PYL-W0212
LOG_FILE = f"{_LABEL}.out"
GEOM_FILE = f"{_LABEL}.xyz"


def base_fn(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    default_inputs: dict[str, Any] | None = None,
    default_blocks: dict[str, Any] | None = None,
    input_swaps: dict[str, Any] | None = None,
    block_swaps: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> cclibSchema:
    """
    Base job function for ORCA recipes.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    default_inputs
        Default input parameters.
    default_blocks
        Default block input parameters.
    input_swaps
        Dictionary of orcasimpleinput swaps for the calculator. To enable new
        entries, set the value as True. To remove entries from the defaults, set
        the value as None.
    block_swaps
        Dictionary of orcablock swaps for the calculator. To enable new entries,
        set the value as True. To remove entries from the defaults, set the
        value as None.
    additional_fields
        Any additional fields to supply to the summarizer.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    cclibSchema
        Dictionary of results from [quacc.schemas.cclib.cclib_summarize_run][]
    """
    inputs = merge_dicts(default_inputs, input_swaps)
    blocks = merge_dicts(default_blocks, block_swaps)
    orcasimpleinput = " ".join(list(inputs.keys()))
    orcablocks = " ".join(list(blocks.keys()))

    atoms.calc = ORCA(
        profile=OrcaProfile(SETTINGS.ORCA_CMD),
        charge=charge,
        mult=spin_multiplicity,
        orcasimpleinput=orcasimpleinput,
        orcablocks=orcablocks,
    )
    atoms = run_calc(atoms, geom_file=GEOM_FILE, copy_files=copy_files)

    return cclib_summarize_run(atoms, LOG_FILE, additional_fields=additional_fields)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from ase.calculators.espresso import Espresso as Espresso_
from ase.calculators.espresso import EspressoProfile
from ase.calculators.espresso import EspressoTemplate as EspressoTemplate_
from ase.io.espresso import construct_namelist

from quacc import SETTINGS
from quacc.calculators.espresso.io import read, write
from quacc.calculators.espresso.keys import ALL_KEYS
from quacc.calculators.espresso.utils import parse_pp_and_cutoff
from quacc.utils.dicts import merge_dicts
from quacc.utils.files import load_yaml_calc

if TYPE_CHECKING:
    from typing import Any

    from ase import Atoms


class EspressoTemplate(EspressoTemplate_):
    """
    This is a wrapper around the ASE Espresso template that allows for the use
    of other binaries such as pw.x, ph.x, cp.x, etc.
    """

    def __init__(self, binary: str = "pw") -> None:
        """
        Initialize the Espresso template.

        Parameters
        ----------
        binary
            The name of the espresso binary to use. This is used to set the
            input/output file names. By default we fall bacl on "pw".

        Returns
        -------
        None
        """
        super().__init__()
        self.inputname = f"{binary}.in"
        self.outputname = f"{binary}.out"
        self.binary = binary

    def write_input(
        self,
        profile: EspressoProfile,
        directory: Path | str,
        atoms: Atoms,
        parameters: dict[str, Any],
        properties: Any,
    ) -> None:
        """
        The function that should be used instead of the one in ASE EspressoTemplate
        to write the input file. It calls a customly defined write function.

        Parameters
        ----------
        profile
            The profile to use.
        directory
            The directory in which to write the input file.
        atoms
            The atoms object to use.
        parameters
            The parameters to use.
        properties
            Special ASE properties

        Returns
        -------
        None
        """
        write(
            Path(directory) / self.inputname,
            atoms,
            binary=self.binary,
            properties=properties,
            pseudo_dir=str(profile.pseudo_path),
            **parameters,
        )

    def read_results(self, directory: Path | str) -> dict[str, Any]:
        """
        The function that should be used instead of the one in ASE EspressoTemplate
        to read the output file. It calls a customly defined read function. It also
        adds the "energy" key to the results dictionnary if it is not present. This
        is needed if the calculation is not made with pw.x.

        Parameters
        ----------
        directory
            The directory in which to read the output file.

        Returns
        -------
        dict
            The results dictionnary
        """
        results = read(Path(directory) / self.outputname, binary=self.binary)
        if "energy" not in results:
            results["energy"] = None
        return results


class Espresso(Espresso_):
    """
    This is a wrapper around the ASE Espresso calculator that adjusts input_data
    parameters and allows for the use of presets. Templates are used to set
    the binary and input/output file names.
    """

    def __init__(
        self,
        input_atoms: Atoms | None = None,
        preset: str | None = None,
        template: EspressoTemplate | None = None,
        profile: EspressoProfile | None = None,
        calc_defaults: dict[str, Any] | None = None,
        parallel_info: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        """
        Initialize the Espresso calculator.

        Parameters
        ----------
        input_atoms
            The input Atoms object to be used for the calculation.
        preset
            The name of a YAML file containing a list of parameters to use as
            a "preset" for the calculator. quacc will automatically look in the
            `ESPRESSO_PRESET_DIR` (default: quacc/calculators/espresso/presets),
            The .yaml extension is not necessary. Any user-supplied calculator
            **kwargs will override any corresponding preset values.
        template
            ASE calculator templace which can be used to specify which espresso
            binary will be used in the calculation. This is taken care of by recipe
            in most cases.
        profile
            ASE calculator profile which can be used to specify the location of
            the espresso binary and pseudopotential files. This is taken care of
            internally using quacc settings.
        calc_defaults
            A dictionary of default input_data parameters to pass to the Espresso
            calculator. These will be overridden by any user-supplied calculator
            **kwargs.
        parallel_info
            parallel_info is a dictionary passed to the ASE Espresso calculator
            profile. It is used to specify prefixes for the command line arguments.
            See the ASE documentation for more details.
        **kwargs
            Additional arguments to be passed to the Espresso calculator, e.g.
            `input_data`, `kpts`... Takes all valid ASE calculator arguments.

        Returns
        -------
        None
        """
        self.preset = preset
        self.input_atoms = input_atoms
        self.calc_defaults = calc_defaults

        template = template or EspressoTemplate("pw")

        kwargs = self._kwargs_handler(template.binary, **kwargs)

        pseudo_path = (
            kwargs["input_data"]
            .get("control", {})
            .get("pseudo_dir", str(SETTINGS.ESPRESSO_PSEUDO))
        )

        bin_path = SETTINGS.ESPRESSO_BIN_PATHS[template.binary]
        profile = profile or EspressoProfile(
            binary=str(bin_path), parallel_info=parallel_info, pseudo_path=pseudo_path
        )

        super().__init__(profile=profile, parallel_info=parallel_info, **kwargs)

        self.template = template

    def _kwargs_handler(self, binary: str, **kwargs) -> dict[str, Any]:
        """
        Function that handles the kwargs. It will merge the user-supplied
        kwargs with the defaults and preset values. Priority order is as follow:

        User-supplied kwargs > preset > defaults

        Parameters
        ----------
        binary
            The espresso binary used to construct the namelist
        **kwargs
            User-supplied kwargs

        Returns
        -------
        kwargs
            The merged kwargs
        """
        keys = ALL_KEYS[binary]
        kwargs["input_data"] = construct_namelist(kwargs.get("input_data"), keys=keys)
        self.calc_defaults["input_data"] = construct_namelist(
            self.calc_defaults["input_data"], keys=keys
        )

        if self.preset:
            config = load_yaml_calc(SETTINGS.ESPRESSO_PRESET_DIR / f"{self.preset}")
            preset_pp = parse_pp_and_cutoff(config, self.input_atoms)
            kwargs = merge_dicts(preset_pp, kwargs)
        return merge_dicts(self.calc_defaults, kwargs)



"""Aliases for type hinting for emmet."""
from __future__ import annotations

from datetime import datetime
from typing import Any, Literal, TypedDict

from emmet.core.math import ListMatrix3D, Matrix3D, Vector3D
from emmet.core.symmetry import CrystalSystem
from emmet.core.vasp.calc_types import CalcType
from emmet.core.vasp.calc_types.enums import RunType, TaskType
from emmet.core.vasp.calculation import VaspObject
from emmet.core.vasp.task_valid import TaskState
from pymatgen.core.composition import Composition
from pymatgen.core.lattice import Lattice
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.io.vasp.inputs import Incar, Kpoints, Poscar, Potcar


class SymmetryData(TypedDict):
    """Type hint associated with `emmet.core.symmetry.SymmetryData`"""

    crystal_system: CrystalSystem
    symbol: str
    number: int
    point_group: str
    symprec: float
    version: str


class PointGroupData(TypedDict):
    """Type hint associated with `emmet.core.symmetry.PointGroupData`"""

    point_group: str
    rotation_number: float
    linear: bool
    tolerance: float
    eigen_tolerance: float
    matrix_tolerance: float


class EmmetBase(TypedDict):
    """Type hint associated with `emmet.core.base.EmmetBaseModel`."""

    emmet_version: str
    pymatgen_version: str
    pull_request: int | None
    database_version: str | None
    build_date: datetime
    license: Literal["BY-C", "BY-NC"]


class StructureMetadata(EmmetBase):
    """Type hint associated with `emmet.core.structure.StructureMetadata`"""

    nsites: int
    elements: list[Element]
    nelements: int
    composition: Composition
    formula_pretty: str
    formula_anonymous: str
    chemsys: str
    volume: float
    density: float
    density_atomic: float
    symmetry: SymmetryData


class MoleculeMetadata(EmmetBase):
    """Type hint associated with `emmet.core.structure.MoleculeMetadata`"""

    charge: int
    spin_multiplicity: int
    natoms: int
    elements: list[Element]
    nelements: int
    nelectrons: int
    composition: Composition
    composition_reduced: Composition
    formula_alphabetical: str
    formula_pretty: str
    formula_anonymous: str
    chemsys: str
    symmetry: PointGroupData


class PotcarSpec(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.PotcarSpec."""

    titel: str
    hash: str


class CalculationInput(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.CalculationInput."""

    incar: dict[str, Any]
    kpoints: dict[str, Any]
    nkpoints: int
    potcar: list[str]
    potcar_spec: list[PotcarSpec]
    potcar_type: list[str]
    parameters: dict[str, Any]
    lattice_rec: Lattice
    structure: Structure
    is_hubbard: bool
    hubbards: dict[str, float]


class ElectronicStep(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.ElectronicStep."""

    alphaZ: float
    ewald: float
    hartreedc: float
    XCdc: float
    pawpsdc: float
    pawaedc: float
    eentropy: float
    bandstr: float
    atom: float
    e_fr_energy: float
    e_wo_entrp: float
    e_0_energy: float


class IonicStep(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.IonicStep."""

    e_fr_energy: float
    e_wo_entrp: float
    e_0_energy: float
    forces: list[Vector3D]
    stress: Matrix3D
    electronic_steps: list[ElectronicStep]
    structure: Structure


class FrequencyDependentDielectric(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.FrequencyDependentDielectric."""

    real: list[list[float]]
    imaginary: list[list[float]]
    energy: list[float]


class ElectronPhononDisplacedStructures(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.ElectronPhononDisplacedStructures."""

    temperatures: list[float]
    structures: list[Structure]


class RunStatistics(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.RunStatistics."""

    average_memory: float
    max_memory: float
    elapsed_time: float
    system_time: float
    user_time: float
    total_time: float
    cores: int


class CalculationOutput(TypedDict, total=False):
    """Type hint associated with emmet.core.vasp.calculation.CalculationOutput."""

    energy: float
    energy_per_atom: float
    structure: Structure
    efermi: float
    is_metal: bool
    bandgap: float
    cbm: float
    vbm: float
    is_gap_direct: bool
    direct_gap: float
    transition: str
    mag_density: float
    epsilon_static: ListMatrix3D
    epsilon_static_wolfe: ListMatrix3D
    epsilon_ionic: ListMatrix3D
    frequency_dependent_dielectric: FrequencyDependentDielectric
    ionic_steps: list[IonicStep]
    locpot: dict[int, list[float]]
    outcar: dict[str, Any]
    force_constants: list[list[Matrix3D]]
    normalmode_frequencies: list[float]
    normalmode_eigenvals: list[float]
    normalmode_eigenvecs: list[list[Vector3D]]
    elph_displaced_structures: ElectronPhononDisplacedStructures
    dos_properties: dict[str, dict[str, dict[str, float]]]
    run_stats: RunStatistics


class Calculation(TypedDict):
    """Type hint associated with emmet.core.vasp.calculation.Calculation."""

    dir_name: str
    vasp_version: str
    has_vasp_completed: TaskState | bool
    input: CalculationInput
    output: CalculationOutput
    completed_at: str
    task_name: str
    output_file_paths: dict[str, str]
    bader: dict
    run_type: RunType
    task_type: TaskType
    calc_type: CalcType


class OrigInputs(TypedDict):
    """Type hint associated with emmet.core.tasks.OrigInputs."""

    incar: Incar
    poscar: Poscar
    kpoints: Kpoints
    potcar: Potcar


class OutputDoc(TypedDict):
    """Type hint associated with emmet.core.tasks.OutputDoc."""

    structure: Structure
    density: float
    energy: float
    forces: list[list[float]]
    stress: list[list[float]]
    energy_per_atom: float
    bandgap: float


class CustodianDoc(TypedDict):
    """Type hint associated with emmet.core.tasks.CustodianDoc."""

    corrections: list[Any]
    job: dict


class AnalysisDoc(TypedDict):
    """Type hint associated with emmet.core.tasks.AnalysisDoc."""

    delta_volume: float
    delta_volume_percent: float
    max_force: float
    warnings: list[str]
    errors: list[str]


class TaskDoc(StructureMetadata):
    """Type hint associated with emmet.core.tasks.TaskDoc."""

    tags: list[str] | None
    dir_name: str
    state: TaskState
    calcs_reversed: list[Calculation]
    structure: Structure
    task_type: CalcType | TaskType
    task_id: str
    orig_inputs: OrigInputs
    output: OutputDoc
    included_objects: VaspObject
    vasp_objects: dict[VaspObject, Any]
    entry: ComputedEntry
    task_label: str
    author: str
    transformations: dict[str, Any]
    additional_json: dict[str, Any]
    custodian: list[CustodianDoc]
    analysis: AnalysisDoc
    last_updated: datetime



"""Core recipes for the Q-Chem."""
from __future__ import annotations

from typing import TYPE_CHECKING

import psutil
from ase.optimize import FIRE

from quacc import SETTINGS, job
from quacc.recipes.qchem._base import base_fn, base_opt_fn

try:
    from sella import Sella

    has_sella = True
except ImportError:
    has_sella = False

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import OptSchema, RunSchema


@job
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    method: str = "wb97mv",
    basis: str = "def2-tzvpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> RunSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-TZVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum implicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to `pymatgen.io.qchem.QChemDictSet` which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """
    defaults = {
        "basis_set": basis,
        "scf_algorithm": scf_algorithm,
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or psutil.cpu_count(logical=False),
        "qchem_input_params": {
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
            "nbo_params": {"version": 7} if SETTINGS.QCHEM_NBO_EXE else None,
        },
    }

    return base_fn(
        atoms,
        charge,
        spin_multiplicity,
        defaults=defaults,
        additional_fields={"name": "Q-Chem Static"},
        copy_files=copy_files,
    )


@job
def freq_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> RunSchema:
    """
    Perform a frequency calculation on a molecular structure.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum implicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to `pymatgen.io.qchem.QChemDictSet` which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """

    defaults = {
        "job_type": "freq",
        "basis_set": basis,
        "scf_algorithm": scf_algorithm,
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or psutil.cpu_count(logical=False),
        "qchem_input_params": {
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    return base_fn(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        defaults=defaults,
        copy_files=copy_files,
        additional_fields={"name": "Q-Chem Frequency"},
    )


@job
def relax_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    method: str = "wb97mv",
    basis: str = "def2-svpd",
    scf_algorithm: str = "diis",
    pcm_dielectric: str | None = None,
    smd_solvent: str | None = None,
    n_cores: int | None = None,
    overwrite_inputs: dict[str, Any] | None = None,
    opt_params: dict[str, Any] | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> OptSchema:
    """
    Optimize aka "relax" a molecular structure with an ASE optimizer.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        DFT exchange-correlation functional or other electronic structure
        method. Defaults to wB97M-V.
    basis
        Basis set. Defaults to def2-SVPD.
    scf_algorithm
        Algorithm used to converge the SCF. Defaults to "diis", but for
        particularly difficult cases, "gdm" should be employed instead.
    pcm_dielectric
        Dielectric constant of the optional polarizable continuum implicit
        solvation model. Defaults to None, in which case PCM will not be
        employed.
    smd_solvent
        Solvent to use for SMD implicit solvation model. Examples include
        "water", "ethanol", "methanol", and "acetonitrile". Refer to the Q-Chem
        manual for a complete list of solvents available. Defaults to None, in
        which case SMD will not be employed.
    n_cores
        Number of cores to use for the Q-Chem calculation. Defaults to use all
        cores available on a given node.
    overwrite_inputs
        Dictionary passed to `pymatgen.io.qchem.QChemDictSet` which can modify
        default values set therein as well as set additional Q-Chem parameters.
        See QChemDictSet documentation for more details.
    opt_params
        Dictionary of custom kwargs for [quacc.runners.ase.run_opt][]
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    OptSchema
        Dictionary of results from [quacc.schemas.ase.summarize_opt_run][]
    """

    calc_defaults = {
        "basis_set": basis,
        "scf_algorithm": scf_algorithm,
        "method": method,
        "charge": charge,
        "spin_multiplicity": spin_multiplicity,
        "cores": n_cores or psutil.cpu_count(logical=False),
        "qchem_input_params": {
            "pcm_dielectric": pcm_dielectric,
            "smd_solvent": smd_solvent,
            "overwrite_inputs": overwrite_inputs,
            "max_scf_cycles": 200 if scf_algorithm.lower() == "gdm" else None,
        },
    }
    opt_defaults = {
        "fmax": 0.01,
        "max_steps": 1000,
        "optimizer": Sella if has_sella else FIRE,
        "optimizer_kwargs": {"use_TRICs": False},
    }

    return base_opt_fn(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        calc_defaults=calc_defaults,
        opt_defaults=opt_defaults,
        opt_params=opt_params,
        additional_fields={"name": "Q-Chem Optimization"},
        copy_files=copy_files,
    )



"""Base jobs for GULP"""
from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

from ase.calculators.gulp import GULP

from quacc import SETTINGS
from quacc.runners.ase import run_calc
from quacc.schemas.ase import summarize_run
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import RunSchema

logger = logging.getLogger(__name__)

GEOM_FILE_PBC = "gulp.cif"
GEOM_FILE_NOPBC = "gulp.xyz"
GULP_CMD = f"{SETTINGS.GULP_CMD} < gulp.gin > gulp.got"


def base_fn(
    atoms: Atoms,
    library: str | None = None,
    keyword_defaults: dict[str, Any] | None = None,
    option_defaults: dict[str, Any] | None = None,
    keyword_swaps: dict[str, Any] | None = None,
    option_swaps: dict[str, Any] | None = None,
    additional_fields: dict[str, Any] | None = None,
) -> RunSchema:
    """
    Base job function for GULP recipes.

    Parameters
    ----------
    atoms
        Atoms object
    library
        Filename of the potential library file, if required.
    keyword_defaults
        Default `keywords` for calculator.
    option_defaults
        Default `options` for calculator.
    keyword_swaps
        Dictionary of custom `keyword` kwargs for the GULP calculator. Set a
        value to `None` to remove a pre-existing key entirely. For a list of
        available keys, refer to the `ase.calculators.gulp.GULP` calculator.
    option_swaps
        Dictionary of custom `options` kwargs for the GULP calculator. Set a
        value to `None` to remove a pre-existing key entirely. For a list of
        available keys, refer to the `ase.calculators.gulp.GULP` calculator.
    additional_fields
        Additional field to supply to the summarizer.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """
    keyword_defaults = keyword_defaults or {}

    if not atoms.pbc.any():
        if keyword_defaults.get("opti") and not keyword_defaults.get("conv"):
            keyword_defaults["conv"] = True
        for k in ["gwolf", "conp"]:
            keyword_defaults.pop(k, None)

    option_defaults = merge_dicts(
        option_defaults,
        {
            f"output cif {GEOM_FILE_PBC}": True if atoms.pbc.any() else None,
            f"output xyz {GEOM_FILE_NOPBC}": None if atoms.pbc.any() else True,
        },
    )

    keywords = merge_dicts(keyword_defaults, keyword_swaps)
    options = merge_dicts(option_defaults, option_swaps)

    gulp_keywords = " ".join(list(keywords.keys()))
    gulp_options = list(options.keys())

    if SETTINGS.GULP_LIB:
        os.environ["GULP_LIB"] = str(SETTINGS.GULP_LIB)
    atoms.calc = GULP(
        command=GULP_CMD, keywords=gulp_keywords, options=gulp_options, library=library
    )
    final_atoms = run_calc(
        atoms, geom_file=GEOM_FILE_PBC if atoms.pbc.any() else GEOM_FILE_NOPBC
    )

    if (
        SETTINGS.CHECK_CONVERGENCE
        and "opti" in gulp_keywords
        and not final_atoms.calc.get_opt_state()
    ):
        msg = "Optimization did not converge."
        raise RuntimeError(msg)

    return summarize_run(
        final_atoms, input_atoms=atoms, additional_fields=additional_fields
    )



"""
Core recipes for EMT.

NOTE: This set of minimal recipes is mainly for demonstration purposes.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ase.calculators.emt import EMT
from ase.optimize import FIRE

from quacc import job
from quacc.runners.ase import run_calc, run_opt
from quacc.schemas.ase import summarize_opt_run, summarize_run
from quacc.utils.dicts import merge_dicts

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import OptSchema, RunSchema


@job
def static_job(atoms: Atoms, **calc_kwargs) -> RunSchema:
    """
    Carry out a static calculation.

    Parameters
    ----------
    atoms
        Atoms object
    **calc_kwargs
        Custom kwargs for the EMT calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `ase.calculators.emt.EMT` calculator.

    Returns
    -------
    RunSchema
        Dictionary of results, specified in [quacc.schemas.ase.summarize_run][]
    """
    atoms.calc = EMT(**calc_kwargs)
    final_atoms = run_calc(atoms)

    return summarize_run(
        final_atoms, input_atoms=atoms, additional_fields={"name": "EMT Static"}
    )


@job
def relax_job(
    atoms: Atoms,
    relax_cell: bool = False,
    opt_params: dict[str, Any] | None = None,
    **calc_kwargs,
) -> OptSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    relax_cell
        Whether to relax the cell
    opt_params
        Dictionary of custom kwargs for the optimization process. Set a value
        to `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to [quacc.runners.ase.run_opt][].
    **calc_kwargs
        Custom kwargs for the EMT calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `ase.calculators.emt.EMT` calculator.

    Returns
    -------
    OptSchema
        Dictionary of results, specified in
        [quacc.schemas.ase.summarize_opt_run][]
    """
    opt_defaults = {"fmax": 0.01, "max_steps": 1000, "optimizer": FIRE}
    opt_flags = merge_dicts(opt_defaults, opt_params)

    atoms.calc = EMT(**calc_kwargs)

    dyn = run_opt(atoms, relax_cell=relax_cell, **opt_flags)

    return summarize_opt_run(dyn, additional_fields={"name": "EMT Relax"})



"""Settings for quacc."""
from __future__ import annotations

import os
from importlib import util
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING, Literal, Optional, Union

from maggma.core import Store
from monty.json import MontyDecoder
from pydantic import Field, field_validator, model_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

if TYPE_CHECKING:
    from typing import Any

installed_engine = next(
    (
        wflow_engine
        for wflow_engine in ["parsl", "covalent", "dask", "prefect", "redun", "jobflow"]
        if util.find_spec(wflow_engine)
    ),
    "local",
)
_DEFAULT_CONFIG_FILE_PATH = Path("~", ".quacc.yaml").expanduser().resolve()


class QuaccSettings(BaseSettings):
    """
    Settings for quacc.

    The default way to modify these is to make a ~/.quacc.yaml file. Alternatively, the
    environment variable QUACC_CONFIG_FILE can be set to point to a custom yaml file
    with quacc settings. The quacc CLI offers a `quacc set <setting> <value>` option to
    do this as well.

    The variables can also be modified individually though environment variables by
    using the "QUACC" prefix. e.g. QUACC_SCRATCH_DIR=/path/to/scratch.
    """

    CONFIG_FILE: Path = Field(
        _DEFAULT_CONFIG_FILE_PATH,
        description=(
            "Path to the YAML file to load alternative quacc configuration "
            "defaults from."
        ),
    )

    # --8<-- [start:settings]

    # ---------------------------
    # Workflow Engine
    # ---------------------------

    WORKFLOW_ENGINE: Literal[
        "covalent", "dask", "parsl", "redun", "jobflow", "prefect", "local"
    ] = Field(
        installed_engine,
        description=(
            "The workflow manager to use."
            "Options include: 'covalent', 'parsl', 'redun', 'jobflow', 'prefect', or 'local'"
        ),
    )

    # ---------------------------
    # General Settings
    # ---------------------------

    RESULTS_DIR: Path = Field(
        Path.cwd(),
        description=(
            "Directory to permanently store I/O-based calculation results in."
            "Note that this behavior may be modified by the chosen workflow engine."
        ),
    )
    SCRATCH_DIR: Optional[Path] = Field(
        None,
        description=(
            "The base directory where calculations are run. If set to None, calculations will be run in a "
            "temporary directory within `RESULTS_DIR`. If a `Path` is supplied, calculations will "
            "be run in a temporary directory within `SCRATCH_DIR`. Files are always moved back "
            "to `RESULTS_DIR` after the calculation is complete, and the temporary directory "
            "in `SCRATCH_DIR` is removed."
        ),
    )
    CREATE_UNIQUE_DIR: bool = Field(
        False,
        description=(
            "Whether to have a unique directory in RESULTS_DIR for each job."
            "Some workflow engines have an option to do this for you already."
        ),
    )
    GZIP_FILES: bool = Field(
        True, description="Whether generated files should be gzip'd."
    )
    CHECK_CONVERGENCE: bool = Field(
        True,
        description="Whether to check for convergence, when implemented by a given recipe.",
    )

    # ---------------------------
    # Data Store Settings
    # ---------------------------
    PRIMARY_STORE: Optional[Union[str, Store]] = Field(
        None,
        description=(
            "String-based JSON representation of the primary Maggma data store "
            "where calculation results will be stored."
            "Taken from the `.to_json()` method of the corresponding Store object."
        ),
    )

    # ---------------------------
    # ORCA Settings
    # ---------------------------
    ORCA_CMD: Path = Field(
        Path(which("orca") or "orca"),
        description=(
            "Path to the ORCA executable. This must be the full, absolute path "
            "for parallel calculations to work."
        ),
    )

    # ---------------------------
    # ESPRESSO Settings
    # ---------------------------
    ESPRESSO_BIN_PATHS: dict[str, Path] = Field(
        {
            "pw": Path("pw.x"),
            "ph": Path("ph.x"),
            "neb": Path("neb.x"),
            "q2r": Path("q2r.x"),
            "matdyn": Path("matdyn.x"),
            "dynmat": Path("dynmat.x"),
            "bands": Path("bands.x"),
            "projwfc": Path("projwfc.x"),
            "pp": Path("pp.x"),
            "wannier90": Path("wannier90.x"),
        },
        description="Name for each espresso binary and its corresponding path. "
        "By default, the binaries are assumed to be in PATH.",
    )
    ESPRESSO_PSEUDO: Optional[Path] = Field(
        None, description=("Path to a pseudopotential library for espresso.")
    )
    ESPRESSO_PRESET_DIR: Path = Field(
        Path(__file__).parent / "calculators" / "espresso" / "presets",
        description="Path to the espresso preset directory",
    )

    # ---------------------------
    # Gaussian Settings
    # ---------------------------
    GAUSSIAN_CMD: Path = Field(
        Path("g16"), description=("Path to the Gaussian executable.")
    )

    # ---------------------------
    # GULP Settings
    # ---------------------------
    GULP_CMD: Path = Field(Path("gulp"), description=("Path to the GULP executable."))
    GULP_LIB: Optional[Path] = Field(
        None,
        description=(
            "Path to the GULP force field library. If not specified, the GULP_LIB environment variable will be used (if present)."
        ),
    )

    # ---------------------------
    # VASP Settings
    # ---------------------------

    # VASP Settings: Main
    VASP_PARALLEL_CMD: str = Field(
        "",
        description=(
            "Parallel command to run VASP with Custodian."
            "For example: srun -N 2 --ntasks-per-node 48"
            "Note that this does not include the executable name."
        ),
    )
    VASP_CMD: str = Field(
        "vasp_std", description="Command to run the standard version of VASP."
    )
    VASP_GAMMA_CMD: str = Field(
        "vasp_gam", description="Command to run the gamma-point only version of VASP."
    )
    VASP_PP_PATH: Optional[Path] = Field(
        None,
        description="Path to the VASP pseudopotential library. Must contain the directories `potpaw_PBE` and `potpaw` for PBE and LDA pseudopotentials, respectively.",
    )
    VASP_VDW: Optional[Path] = Field(
        None, description="Path to the vdw_kernel.bindat file for VASP vdW functionals."
    )

    # VASP Settings: General
    VASP_INCAR_COPILOT: Literal["off", "on", "aggressive"] = Field(
        "on",
        description=(
            "Controls VASP co-pilot mode for automated INCAR parameter handling."
            "off: Do not use co-pilot mode. INCAR parameters will be unmodified."
            "on: Use co-pilot mode. This will only modify INCAR flags not already set by the user."
            "aggressive: Use co-pilot mode in aggressive mode. This will modify INCAR flags even if they are already set by the user."
        ),
    )
    VASP_BADER: bool = Field(
        bool(which("bader")),
        description=(
            "Whether to run a Bader analysis when summarizing VASP results."
            "Requires bader to be in PATH."
        ),
    )
    VASP_CHARGEMOL: bool = Field(
        bool(os.environ.get("DDEC6_ATOMIC_DENSITIES_DIR")),
        description=(
            "Whether to run a Chargemol (i.e. DDEC6, CM5) analysis when summarizing VASP results."
            "Requires the Chargemol executable to be in PATH and the DDEC6_ATOMIC_DENSITIES_DIR environment variable."
        ),
    )
    VASP_PRESET_MAG_DEFAULT: float = Field(
        1.0,
        description=(
            "Default initial magmom to use for a given element if a preset "
            "with magmoms is provided but an element is missing from the list"
        ),
    )
    VASP_MAG_CUTOFF: float = Field(
        0.05,
        description=(
            "If the absolute value of all magnetic moments are below this value, "
            "they will be set to 0 such that a spin-unpolarized calculation will be performed"
        ),
    )
    VASP_COPY_MAGMOMS: bool = Field(
        True,
        description=(
            "If True, any pre-existing atoms.get_magnetic_moments() will be set"
            "in atoms.set_initial_magnetic_moments()."
        ),
    )
    VASP_PRESET_DIR: Path = Field(
        Path(__file__).parent / "calculators" / "vasp" / "presets",
        description="Path to the VASP preset directory",
    )

    # VASP Settings: Custodian
    VASP_USE_CUSTODIAN: bool = Field(
        True, description="Whether Custodian should be used to run VASP"
    )
    VASP_CUSTODIAN_VTST: bool = Field(
        False,
        description=(
            "If VTST-related input swaps should be used when running Custodian."
            "Requires VASP to be compiled with VTST"
        ),
    )
    VASP_CUSTODIAN_MAX_ERRORS: int = Field(
        5, description="Maximum errors for Custodian"
    )
    VASP_CUSTODIAN_HANDLERS: list[str] = Field(
        [
            "VaspErrorHandler",
            "MeshSymmetryErrorHandler",
            "UnconvergedErrorHandler",
            "NonConvergingErrorHandler",
            "PotimErrorHandler",
            "PositiveEnergyErrorHandler",
            "FrozenJobErrorHandler",
            "StdErrHandler",
            "LargeSigmaHandler",
            "IncorrectSmearingHandler",
        ],
        description="Handlers for Custodian",
    )
    VASP_CUSTODIAN_VALIDATORS: list[str] = Field(
        ["VasprunXMLValidator", "VaspFilesValidator"],
        description="Validators for Custodian",
    )
    VASP_CUSTODIAN_WALL_TIME: Optional[int] = Field(
        None,
        description=(
            "After this many seconds, Custodian will stop running "
            "and ensure that VASP writes a STOPCAR"
        ),
    )

    # ---------------------------
    # Q-Chem Settings
    # ---------------------------

    # Q-Chem Settings: Main
    QCHEM_CMD: str = Field(
        "qchem", description="Command to run the standard version of Q-Chem."
    )

    QCHEM_LOCAL_SCRATCH: Path = Field(
        Path("/tmp") if Path("/tmp").exists() else Path.cwd() / ".qchem_scratch",
        description="Compute-node local scratch directory in which Q-Chem should perform IO.",
    )

    # Q-Chem Settings: Custodian
    QCHEM_USE_ERROR_HANDLERS: bool = Field(
        True,
        description="Whether Custodian's error handlers should be employed for Q-Chem.",
    )

    QCHEM_CUSTODIAN_MAX_ERRORS: int = Field(
        5, description="Maximum errors for Q-Chem Custodian."
    )

    # NBO Settings
    QCHEM_NBO_EXE: Optional[Path] = Field(
        None, description="Full path to the NBO executable."
    )

    # ---------------------------
    # NewtonNet Settings
    # ---------------------------
    NEWTONNET_MODEL_PATH: Union[Path, list[Path]] = Field(
        "best_model_state.tar", description="Path to NewtonNet .tar model"
    )
    NEWTONNET_CONFIG_PATH: Union[Path, list[Path]] = Field(
        "config.yml", description="Path to NewtonNet YAML settings file"
    )

    # --8<-- [end:settings]

    @field_validator("RESULTS_DIR", "SCRATCH_DIR")
    @classmethod
    def resolve_and_make_paths(cls, v: Optional[Path]) -> Optional[Path]:
        """Resolve and make paths."""
        if v is None:
            return v

        v = Path(os.path.expandvars(v)).expanduser().resolve()
        if not v.exists():
            os.makedirs(v)
        return v

    @field_validator(
        "ESPRESSO_PRESET_DIR",
        "ESPRESSO_PSEUDO",
        "GAUSSIAN_CMD",
        "GULP_CMD",
        "GULP_LIB",
        "ORCA_CMD",
        "QCHEM_LOCAL_SCRATCH",
        "NEWTONNET_MODEL_PATH",
        "VASP_PRESET_DIR",
        "VASP_PP_PATH",
        "VASP_VDW",
    )
    @classmethod
    def expand_paths(cls, v: Optional[Path]) -> Optional[Path]:
        """Expand ~/ in paths."""
        return v.expanduser() if v is not None else v

    @field_validator("PRIMARY_STORE")
    def generate_store(cls, v: Union[str, Store]) -> Store:
        """Generate the Maggma store"""
        return MontyDecoder().decode(v) if isinstance(v, str) else v

    model_config = SettingsConfigDict(env_prefix="quacc_")

    @model_validator(mode="before")
    @classmethod
    def load_default_settings(cls, values: dict[str, Any]) -> dict[str, Any]:
        """
        Loads settings from a root file if available and uses that as defaults in place
        of built in defaults.

        Parameters
        ----------
        values
            Settings to load.

        Returns
        -------
        dict
            Loaded settings.
        """

        from monty.serialization import loadfn

        config_file_path = (
            Path(values.get("CONFIG_FILE", _DEFAULT_CONFIG_FILE_PATH))
            .expanduser()
            .resolve()
        )

        new_values = {}  # type: dict
        if config_file_path.exists() and config_file_path.stat().st_size > 0:
            new_values |= loadfn(config_file_path)

        new_values.update(values)
        return new_values



"""Schemas for storing ASE-based data."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ase import units
from ase.filters import Filter
from ase.io import read
from ase.vibrations.data import VibrationsData

from quacc import SETTINGS
from quacc.schemas.atoms import atoms_to_metadata
from quacc.schemas.prep import prep_next_run as prep_next_run_
from quacc.utils.dicts import merge_several_dicts, sort_dict
from quacc.utils.files import get_uri
from quacc.wflow_tools.db import results_to_db

if TYPE_CHECKING:
    from typing import Any

    from ase.atoms import Atoms
    from ase.io import Trajectory
    from ase.optimize.optimize import Optimizer
    from ase.thermochemistry import IdealGasThermo
    from ase.vibrations import Vibrations
    from maggma.core import Store

    from quacc.schemas._aliases.ase import (
        OptSchema,
        RunSchema,
        ThermoSchema,
        VibSchema,
        VibThermoSchema,
    )


def summarize_run(
    atoms: Atoms,
    input_atoms: Atoms | None = None,
    charge_and_multiplicity: tuple[int, int] | None = None,
    prep_next_run: bool = True,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> RunSchema:
    """
    Get tabulated results from an Atoms object and calculator and store them in a
    database-friendly format. This is meant to be compatible with all calculator types.

    Parameters
    ----------
    atoms
        ASE Atoms following a calculation. A calculator must be attached.
    input_atoms
        Input ASE Atoms object to store.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    prep_next_run
        Whether the Atoms object stored in `{"atoms": atoms}` should be prepared
        for the next run. This clears out any attached calculator and moves the
        final magmoms to the initial magmoms.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    RunSchema
        Dictionary representation of the task document
    """

    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    if not atoms.calc:
        msg = "ASE Atoms object has no attached calculator."
        raise ValueError(msg)
    if not atoms.calc.results:
        msg = "ASE Atoms object's calculator has no results."
        raise ValueError(msg)

    uri = get_uri(Path.cwd())
    input_atoms_metadata = (
        atoms_to_metadata(input_atoms, charge_and_multiplicity=charge_and_multiplicity)
        if input_atoms
        else None
    )
    inputs = {
        "parameters": atoms.calc.parameters,
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
        "input_atoms": input_atoms_metadata,
    }

    results = {"results": atoms.calc.results}

    atoms_to_store = prep_next_run_(atoms) if prep_next_run else atoms

    if atoms:
        final_atoms_metadata = atoms_to_metadata(
            atoms_to_store, charge_and_multiplicity=charge_and_multiplicity
        )
    else:
        final_atoms_metadata = {}

    unsorted_task_doc = merge_several_dicts(
        final_atoms_metadata, inputs, results, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_opt_run(
    dyn: Optimizer,
    trajectory: Trajectory | list[Atoms] = None,
    check_convergence: bool | None = None,
    charge_and_multiplicity: tuple[int, int] | None = None,
    prep_next_run: bool = True,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> OptSchema:
    """
    Get tabulated results from an ASE Atoms trajectory and store them in a database-
    friendly format. This is meant to be compatible with all calculator types.

    Parameters
    ----------
    dyn
        ASE Optimizer object.
    trajectory
        ASE Trajectory object or list[Atoms] from reading a trajectory file. If
        None, the trajectory must be found in dyn.traj_atoms.
    check_convergence
        Whether to check the convergence of the calculation. Defaults to True in
        settings.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    prep_next_run
        Whether the Atoms object stored in {"atoms": atoms} should be prepared
        for the next run This clears out any attached calculator and moves the
        final magmoms to the initial magmoms.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    OptSchema
        Dictionary representation of the task document
    """

    check_convergence = (
        SETTINGS.CHECK_CONVERGENCE if check_convergence is None else check_convergence
    )
    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    # Check convergence
    is_converged = dyn.converged()
    if check_convergence and not is_converged:
        msg = f"Optimization did not converge. Refer to {Path.cwd()}"
        raise RuntimeError(msg)

    # Get trajectory
    if not trajectory:
        trajectory = (
            dyn.traj_atoms
            if hasattr(dyn, "traj_atoms")
            else read(dyn.trajectory.filename, index=":")
        )

    initial_atoms = trajectory[0]
    final_atoms = dyn.atoms.atoms if isinstance(dyn.atoms, Filter) else dyn.atoms

    # Base task doc
    base_task_doc = summarize_run(
        final_atoms,
        input_atoms=initial_atoms,
        charge_and_multiplicity=charge_and_multiplicity,
        prep_next_run=prep_next_run,
        store=False,
    )

    opt_fields = {
        "fmax": getattr(dyn, "fmax", None),
        "parameters_opt": dyn.todict(),
        "converged": is_converged,
        "nsteps": dyn.get_number_of_steps(),
        "trajectory": [
            atoms_to_metadata(atoms, charge_and_multiplicity=charge_and_multiplicity)
            for atoms in trajectory
        ],
        "trajectory_results": [atoms.calc.results for atoms in trajectory],
    }

    # Create a dictionary of the inputs/outputs
    unsorted_task_doc = merge_several_dicts(
        base_task_doc, opt_fields, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_vib_run(
    vib: Vibrations,
    charge_and_multiplicity: tuple[int, int] | None = None,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> VibSchema:
    """
    Get tabulated results from an ASE Vibrations object and store them in a database-
    friendly format.

    Parameters
    ----------
    vib
        ASE Vibrations object.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    VibSchema
        Dictionary representation of the task document
    """
    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    vib_freqs_raw = vib.get_frequencies().tolist()
    vib_energies_raw = vib.get_energies().tolist()

    # Convert imaginary modes to negative values for DB storage
    for i, f in enumerate(vib_freqs_raw):
        if np.imag(f) > 0:
            vib_freqs_raw[i] = -np.abs(f)
            vib_energies_raw[i] = -np.abs(vib_energies_raw[i])
        else:
            vib_freqs_raw[i] = np.abs(f)
            vib_energies_raw[i] = np.abs(vib_energies_raw[i])

    atoms = vib._atoms if isinstance(vib, VibrationsData) else vib.atoms

    uri = get_uri(Path.cwd())
    inputs = {
        "parameters": None
        if isinstance(vib, VibrationsData)
        else atoms.calc.parameters,
        "parameters_vib": None
        if isinstance(vib, VibrationsData)
        else {
            "delta": vib.delta,
            "direction": vib.direction,
            "method": vib.method,
            "ndof": vib.ndof,
            "nfree": vib.nfree,
        },
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }

    atoms_metadata = atoms_to_metadata(
        atoms, charge_and_multiplicity=charge_and_multiplicity
    )

    # Get the true vibrational modes
    natoms = len(atoms)
    if natoms == 1:
        vib_freqs = []
        vib_energies = []
    elif atoms.pbc.any():
        vib_freqs = vib_freqs_raw
        vib_energies = vib_energies_raw
    else:
        # Sort by absolute value
        vib_freqs_raw_sorted = vib_freqs_raw.copy()
        vib_energies_raw_sorted = vib_energies_raw.copy()
        vib_freqs_raw_sorted.sort(key=np.abs)
        vib_energies_raw_sorted.sort(key=np.abs)

        # Cut the 3N-5 or 3N-6 modes based on their absolute value
        n_modes = (
            3 * natoms - 5 if atoms_metadata["symmetry"]["linear"] else 3 * natoms - 6
        )
        vib_freqs = vib_freqs_raw_sorted[-n_modes:]
        vib_energies = vib_energies_raw_sorted[-n_modes:]

    imag_vib_freqs = [f for f in vib_freqs if f < 0]

    results = {
        "results": {
            "imag_vib_freqs": imag_vib_freqs,
            "n_imag": len(imag_vib_freqs),
            "vib_energies": vib_energies,
            "vib_freqs": vib_freqs,
            "vib_energies_raw": vib_energies_raw,
            "vib_freqs_raw": vib_freqs_raw,
        }
    }

    unsorted_task_doc = merge_several_dicts(
        atoms_metadata, inputs, results, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_ideal_gas_thermo(
    igt: IdealGasThermo,
    temperature: float = 298.15,
    pressure: float = 1.0,
    charge_and_multiplicity: tuple[int, int] | None = None,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> ThermoSchema:
    """
    Get tabulated results from an ASE IdealGasThermo object and store them in a
    database-friendly format.

    Parameters
    ----------
    igt
        ASE IdealGasThermo object.
    temperature
        Temperature in Kelvins.
    pressure
        Pressure in bar.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    ThermoSchema
        Dictionary representation of the task document
    """

    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    uri = get_uri(Path.cwd())
    spin_multiplicity = round(2 * igt.spin + 1)

    inputs = {
        "parameters_thermo": {
            "temperature": temperature,
            "pressure": pressure,
            "sigma": igt.sigma,
            "spin_multiplicity": spin_multiplicity,
            "vib_freqs": [e / units.invcm for e in igt.vib_energies],
            "vib_energies": igt.vib_energies.tolist(),
            "n_imag": igt.n_imag,
        },
        "nid": uri.split(":")[0],
        "dir_name": ":".join(uri.split(":")[1:]),
    }

    results = {
        "results": {
            "energy": igt.potentialenergy,
            "enthalpy": igt.get_enthalpy(temperature),
            "entropy": igt.get_entropy(temperature, pressure * 10**5),
            "gibbs_energy": igt.get_gibbs_energy(temperature, pressure * 10**5),
            "zpe": igt.get_ZPE_correction(),
        }
    }

    if charge_and_multiplicity and spin_multiplicity != charge_and_multiplicity[1]:
        msg = (
            "The IdealGasThermo spin multiplicity does not match the user-specified multiplicity.",
        )
        raise ValueError(msg)

    atoms_metadata = atoms_to_metadata(
        igt.atoms, charge_and_multiplicity=charge_and_multiplicity
    )

    unsorted_task_doc = merge_several_dicts(
        atoms_metadata, inputs, results, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc


def summarize_vib_and_thermo(
    vib: Vibrations,
    igt: IdealGasThermo,
    temperature: float = 298.15,
    pressure: float = 1.0,
    charge_and_multiplicity: tuple[int, int] | None = None,
    additional_fields: dict[str, Any] | None = None,
    store: Store | bool | None = None,
) -> VibThermoSchema:
    """
    Get tabulated results from an ASE Vibrations run and ASE IdealGasThermo object and
    store them in a database-friendly format.

    Parameters
    ----------
    vib
        ASE Vibrations object.
    igt
        ASE IdealGasThermo object.
    temperature
        Temperature in Kelvins.
    pressure
        Pressure in bar.
    charge_and_multiplicity
        Charge and spin multiplicity of the Atoms object, only used for Molecule
        metadata.
    additional_fields
        Additional fields to add to the task document.
    store
        Maggma Store object to store the results in. If None,
        `SETTINGS.PRIMARY_STORE` will be used.

    Returns
    -------
    VibThermoSchema
        A dictionary that merges the `VibSchema` and `ThermoSchema`.
    """
    additional_fields = additional_fields or {}
    store = SETTINGS.PRIMARY_STORE if store is None else store

    vib_task_doc = summarize_vib_run(
        vib, charge_and_multiplicity=charge_and_multiplicity, store=False
    )
    thermo_task_doc = summarize_ideal_gas_thermo(
        igt,
        temperature=temperature,
        pressure=pressure,
        charge_and_multiplicity=charge_and_multiplicity,
        store=False,
    )

    unsorted_task_doc = merge_several_dicts(
        vib_task_doc, thermo_task_doc, additional_fields
    )
    task_doc = sort_dict(unsorted_task_doc)

    if store:
        results_to_db(store, task_doc)

    return task_doc



"""I/O utilities for the espresso calculator."""
from __future__ import annotations

import re
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
from ase import Atoms
from ase.io import read as _read
from ase.io import write as _write
from ase.io.espresso import namelist_to_string
from ase.units import Bohr

from quacc import SETTINGS

if TYPE_CHECKING:
    from typing import Any, TextIO

freg = re.compile(r"-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+\-]?\d+)?")


def write(
    filename: str | Path,
    atoms: Atoms,
    binary: str = "pw",
    properties: dict[str, Any] | None = None,
    **kwargs,
) -> None:
    """
    Wrapper around ase.io.espresso.write to read additional binary, such as ph.x.
    This is probably temporary until ase.io.espresso is updated.
    """

    filename_path = Path(filename).expanduser().resolve()
    if SETTINGS.SCRATCH_DIR not in filename_path.parents:
        raise ValueError(f"File {filename_path} is not in the scratch directory.")
    with Path.open(filename, "w") as fd:
        write_espresso_dict[binary](fd, atoms=atoms, properties=properties, **kwargs)


def read(filename: str | Path, binary: str = "pw") -> dict[str, Any]:
    """
    Wrapper around ase.io.espresso.read to read additional binary, such as ph.x.
    This is probably temporary until ase.io.espresso is updated.
    """

    filename_path = Path(filename).expanduser().resolve()
    if SETTINGS.SCRATCH_DIR not in filename_path.parents:
        raise ValueError(f"File {filename_path} is not in the scratch directory.")
    with Path.open(filename) as fd:
        return read_espresso_dict[binary](fd)


def write_espresso_io(fd: TextIO, **kwargs) -> None:
    """
    Function which writes input for simple espresso binaries without special cards.
    Non-exhaustive list (to complete):

    - pp.x
    - dynmat.x
    - projwfc.x

    Note: "EOF" is appended at the end of the file.
    (https://lists.quantum-espresso.org/pipermail/users/2020-November/046269.html)

    Parameters
    ----------
    fd
        The file descriptor of the input file.

    kwargs
        kwargs dictionary possibly containing the following keys:

        - input_data: dict

    Returns
    -------
    None
    """

    pwi = namelist_to_string(kwargs["input_data"])[:-1]
    pwi.append("EOF")
    fd.write("".join(pwi))


def write_espresso_ph(fd: TextIO, **kwargs) -> None:
    """
    Function that write the input file for a ph.x calculation. Normal namelist
    cards are passed in the input_data dictionary. Which can be either nested
    or flat, ASE style. The q-points are passed in the qpts list. If qplot is
    set to True then qpts is expected to be a list of list|tuple of length 4.
    Where the first three elements are the coordinates of the q-point in units
    of 2pi/alat and the last element is the weight of the q-point. if qplot is
    set to False then qpts is expected to be a simple list of length 4 (single
    q-point). Finally if ldisp is set to True, the above is discarded and the
    q-points are read from the nq1, nq2, nq3 cards in the input_data dictionary.

    Additionally, a nat_todo kwargs (list[int]) can be specified in the kwargs.
    It will be used if nat_todo is set to True in the input_data dictionary.

    Globally, this function follows the convention set in the ph.x documentation
    (https://www.quantum-espresso.org/Doc/INPUT_PH.html)

    Parameters
    ----------
    fd
        The file descriptor of the input file.

    kwargs
        kwargs dictionary possibly containing the following keys:

        - input_data: dict
        - qpts: list[list[float]] | list[tuple[float]] | list[float]
        - nat_todo: list[int]

    Returns
    -------
    None
    """

    input_data = kwargs.get("input_data")
    pwi = namelist_to_string(input_data)[:-1]
    fd.write("".join(pwi))

    qpts = kwargs.get("qpts")
    qplot = input_data["inputph"].get("qplot", False)
    ldisp = input_data["inputph"].get("ldisp", False)
    nat_todo = input_data["inputph"].get("nat_todo", 0)

    if qplot and ldisp:
        fd.write(f"{len(qpts)}\n")
        for qpt in qpts:
            fd.write(f"{qpt[0]:0.8f} {qpt[1]:0.8f} {qpt[2]:0.8f} {qpt[3]:1d}\n")
    else:
        fd.write(f"{qpts[0]:0.8f} {qpts[1]:0.8f} {qpts[2]:0.8f}\n")
    if nat_todo:
        nat_todo_index = kwargs.get("nat_todo")
        tmp = [str(i) for i in nat_todo_index]
        fd.write(" ".join(tmp))
        fd.write("\n")


def read_espresso_ph(fd: TextIO) -> dict[str, Any]:
    """
    Function that reads the output of a ph.x calculation. It returns a dictionary
    where each q-point is a key and the value is a dictionary with the following
    keys if available:

    - qpoints: The q-point in cartesian coordinates.
    - kpoints: The k-points in cartesian coordinates.
    - dieltensor: The dielectric tensor.
    - borneffcharge: The effective Born charges.
    - borneffcharge_dfpt: The effective Born charges from DFPT.
    - polarizability: The polarizability tensor.
    - modes: The phonon modes.
    - eqpoints: The symmetrically equivalent q-points.
    - freqs: The phonon frequencies.
    - mode_symmetries: The symmetries of the modes.
    - atoms: The atoms object.

    Some notes:

        - For some reason, the cell is not defined to high level of precision with
            ph.x. Be careful when using the atoms object retrieved from this function.
        - This function can be called on incomplete calculations i.e. if the calculation
            couldn't diagonalize the dynamical matrix for some q-points, the results for
            the other q-points will still be returned.

    Parameters
    ----------
    fd
        The file descriptor of the output file.

    Returns
    -------
    dict
        The results dictionnary as described above.
    """
    QPOINTS = r"(?i)^\s*Calculation\s*of\s*q"
    NKPTS = r"(?i)^\s*number\s*of\s*k\s*points\s*"
    DIEL = r"(?i)^\s*Dielectric\s*constant\s*in\s*cartesian\s*axis\s*$"
    BORN = r"(?i)^\s*Effective\s*charges\s*\(d\s*Force\s*/\s*dE\)"
    POLA = r"(?i)^\s*Polarizability\s*(a.u.)\^3"
    MODE = r"(?i)^\s*(mode\s*#\s*\d\s*)+"
    EQPOINTS = r"(?i)^\s*Number\s*of\s*q\s*in\s*the\s*star\s*=\s*"
    DIAG = r"(?i)^\s*Diagonalizing\s*the\s*dynamical\s*matrix\s*$"
    MODE_SYM = r"(?i)^\s*Mode\s*symmetry,\s*"
    BORN_DFPT = r"(?i)^\s*Effective\s*charges\s*\(d\s*P\s*/\s*du\)"
    POSITIONS = r"(?i)^\s*site\s*n\..*\(alat\s*units\)"
    ALAT = r"(?i)^\s*celldm\(1\)="
    CELL = r"^\s*crystal\s*axes:\s*\(cart.\s*coord.\s*in\s*units\s*of\s*alat\)"
    ELECTRON_PHONON = r"(?i)^\s*electron-phonon\s*interaction\s*...\s*$"

    output = {
        QPOINTS: [],
        NKPTS: [],
        DIEL: [],
        BORN: [],
        BORN_DFPT: [],
        POLA: [],
        MODE: [],
        EQPOINTS: [],
        DIAG: [],
        MODE_SYM: [],
        POSITIONS: [],
        ALAT: [],
        CELL: [],
        ELECTRON_PHONON: [],
    }

    names = {
        QPOINTS: "qpoints",
        NKPTS: "kpoints",
        DIEL: "dieltensor",
        BORN: "borneffcharge",
        BORN_DFPT: "borneffcharge_dfpt",
        POLA: "polarizability",
        MODE: "modes",
        EQPOINTS: "eqpoints",
        DIAG: "freqs",
        MODE_SYM: "mode_symmetries",
        POSITIONS: "positions",
        ALAT: "alat",
        CELL: "cell",
        ELECTRON_PHONON: "ep_data",
    }

    unique = {
        QPOINTS: True,
        NKPTS: False,
        DIEL: True,
        BORN: True,
        BORN_DFPT: True,
        POLA: True,
        MODE: False,
        EQPOINTS: True,
        DIAG: True,
        MODE_SYM: True,
        POSITIONS: True,
        ALAT: True,
        CELL: True,
        ELECTRON_PHONON: True,
    }

    results = {}
    fdo_lines = [i for i in fd.read().splitlines() if i]
    n_lines = len(fdo_lines)

    for idx, line in enumerate(fdo_lines):
        for key in output:
            if bool(re.match(key, line)):
                output[key].append(idx)

    output = {key: np.array(value) for key, value in output.items()}

    def _read_qpoints(idx):
        match = re.findall(freg, fdo_lines[idx])
        return tuple(float(x) for x in match)

    def _read_kpoints(idx):
        n_kpts = int(re.findall(freg, fdo_lines[idx])[0])
        kpts = []
        for line in fdo_lines[idx + 2 : idx + 2 + n_kpts]:
            if bool(re.search(r"^\s*k\(.*wk", line)):
                kpts.append([float(x) for x in re.findall(freg, line)[1:]])
        return np.array(kpts)

    def _read_modes(idx):
        n = 1
        n_modes = len(re.findall(r"mode", fdo_lines[idx]))
        modes = []
        while not modes or bool(re.match(r"^\s*\(", fdo_lines[idx + n])):
            tmp = re.findall(freg, fdo_lines[idx + n])
            modes.append([float(x) for x in tmp])
            n += 1
        return np.hsplit(np.array(modes), n_modes)

    def _read_eqpoints(idx):
        n_star = int(re.findall(freg, fdo_lines[idx])[0])
        return np.loadtxt(
            fdo_lines[idx + 2 : idx + 2 + n_star], usecols=(1, 2, 3)
        ).reshape(-1, 3)

    def _read_freqs(idx):
        n = 0
        freqs = []
        stop = 0
        while not freqs or stop < 2:
            if bool(re.search(r"^\s*freq", fdo_lines[idx + n])):
                tmp = re.findall(freg, fdo_lines[idx + n])[1]
                freqs.append(float(tmp))
            if bool(re.search(r"\*{5,}", fdo_lines[idx + n])):
                stop += 1
            n += 1
        return np.array(freqs)

    def _read_sym(idx):
        n = 1
        sym = {}
        while bool(re.match(r"^\s*freq", fdo_lines[idx + n])):
            r = re.findall("\\d+", fdo_lines[idx + n])
            r = tuple(range(int(r[0]), int(r[1]) + 1))
            sym[r] = fdo_lines[idx + n].split("-->")[1].strip()
            sym[r] = re.sub(r"\s+", " ", sym[r])
            n += 1
        return sym

    def _read_epsil(idx):
        epsil = np.zeros((3, 3))
        for n in range(1, 4):
            tmp = re.findall(freg, fdo_lines[idx + n])
            epsil[n - 1] = [float(x) for x in tmp]
        return epsil

    def _read_born(idx):
        n = 1
        born = []
        while idx + n < n_lines:
            if re.search(r"^\s*atom\s*\d\s*\S", fdo_lines[idx + n]):
                pass
            elif re.search(r"^\s*E(x|y|z)\s*\(", fdo_lines[idx + n]):
                tmp = re.findall(freg, fdo_lines[idx + n])
                born.append([float(x) for x in tmp])
            else:
                break
            n += 1
        born = np.array(born)
        return np.vsplit(born, len(born) // 3)

    def _read_born_dfpt(idx):
        n = 1
        born = []
        while idx + n < n_lines:
            if re.search(r"^\s*atom\s*\d\s*\S", fdo_lines[idx + n]):
                pass
            elif re.search(r"^\s*P(x|y|z)\s*\(", fdo_lines[idx + n]):
                tmp = re.findall(freg, fdo_lines[idx + n])
                born.append([float(x) for x in tmp])
            else:
                break
            n += 1
        born = np.array(born)
        return np.vsplit(born, len(born) // 3)

    def _read_pola(idx):
        pola = np.zeros((3, 3))
        for n in range(1, 4):
            tmp = re.findall(freg, fdo_lines[idx + n])[:3]
            pola[n - 1] = [float(x) for x in tmp]
        return pola

    def _read_positions(idx):
        positions = []
        symbols = []
        n = 1
        while re.findall(r"^\s*\d+", fdo_lines[idx + n]):
            symbols.append(fdo_lines[idx + n].split()[1])
            positions.append(
                [float(x) for x in re.findall(freg, fdo_lines[idx + n])[-3:]]
            )
            n += 1
        atoms = Atoms(positions=positions, symbols=symbols)
        atoms.pbc = True
        return atoms

    def _read_alat(idx):
        return float(re.findall(freg, fdo_lines[idx])[1])

    def _read_cell(idx):
        cell = []
        n = 1
        while re.findall(r"^\s*a\(\d\)", fdo_lines[idx + n]):
            cell.append([float(x) for x in re.findall(freg, fdo_lines[idx + n])[-3:]])
            n += 1
        return np.array(cell)

    def _read_electron_phonon(idx):
        results = {}

        broad_re = r"^\s*Gaussian\s*Broadening:\s+([\d.]+)\s+Ry, ngauss=\s+\d+"
        dos_re = r"^\s*DOS\s*=\s*([\d.]+)\s*states/spin/Ry/Unit\s*Cell\s*at\s*Ef=\s+([\d.]+)\s+eV"
        lg_re = r"^\s*lambda\(\s+(\d+)\)=\s+([\d.]+)\s+gamma=\s+([\d.]+)\s+GHz"
        end_re = r"^\s*Number\s*of\s*q\s*in\s*the\s*star\s*=\s+(\d+)$"

        lambdas = []
        gammas = []

        current = None

        n = 1
        while idx + n < n_lines:
            line = fdo_lines[idx + n]

            broad_match = re.match(broad_re, line)
            dos_match = re.match(dos_re, line)
            lg_match = re.match(lg_re, line)
            end_match = re.match(end_re, line)

            if broad_match:
                if lambdas:
                    results[current]["lambdas"] = lambdas
                    results[current]["gammas"] = gammas
                    lambdas = []
                    gammas = []
                current = broad_match[1]
                results[current] = {}
            elif dos_match:
                results[current]["dos"] = float(dos_match[1])
                results[current]["fermi"] = float(dos_match[2])
            elif lg_match:
                lambdas.append(float(lg_match[2]))
                gammas.append(float(lg_match[3]))

            if end_match:
                results[current]["lambdas"] = lambdas
                results[current]["gammas"] = gammas
                break

            n += 1

        return results

    properties = {
        NKPTS: _read_kpoints,
        DIEL: _read_epsil,
        BORN: _read_born,
        BORN_DFPT: _read_born_dfpt,
        POLA: _read_pola,
        MODE: _read_modes,
        EQPOINTS: _read_eqpoints,
        DIAG: _read_freqs,
        MODE_SYM: _read_sym,
        POSITIONS: _read_positions,
        ALAT: _read_alat,
        CELL: _read_cell,
        ELECTRON_PHONON: _read_electron_phonon,
    }

    iblocks = np.append(output[QPOINTS], n_lines)

    for past, future in zip(iblocks[:-1], iblocks[1:]):
        qpoint = _read_qpoints(past)
        results[qpoint] = {}
        for prop in properties:
            p = (past < output[prop]) & (output[prop] < future)
            selected = output[prop][p]
            if len(selected) == 0:
                continue
            if unique[prop]:
                idx = output[prop][p][-1]
                results[qpoint][names[prop]] = properties[prop](idx)
            else:
                tmp = {k + 1: 0 for k in range(len(selected))}
                for k, idx in enumerate(selected):
                    tmp[k + 1] = properties[prop](idx)
                results[qpoint][names[prop]] = tmp
        alat = results[qpoint].pop("alat", 1.0)
        atoms = results[qpoint].pop("positions", None)
        cell = results[qpoint].pop("cell", np.eye(3))
        if atoms:
            atoms.positions *= alat * Bohr
            atoms.cell = cell * alat * Bohr
            atoms.wrap()
            results[qpoint]["atoms"] = atoms

    return results


def read_espresso_pw(filename: str | Path) -> dict[str, Any]:
    """
    Wrapper around ase.io.espresso.read_espresso_out to read the output of a pw.x
    and directly return the dictionary of properties to quacc.

    Parameters
    ----------
    filename
        The name of the output file.

    Returns
    -------
    dict
        The results dictionary.
    """
    atoms = _read(filename, format="espresso-out")
    return dict(atoms.calc.properties())


def write_espresso_pw(
    filename: str | Path, atoms: Atoms, properties: dict[str, Any] | None, **kwargs
) -> None:
    """
    Wrapper around ase.io.espresso.write_espresso_in to write the input of a pw.x

    Parameters
    ----------
    filename
        The name of the output file.

    atoms
        The atoms object.

    properties
        Additional ASE properties

    kwargs
        Additional keyword arguments to pass to ase.io.espresso.write_espresso_in

    Returns
    -------
    None
    """
    _write(filename, atoms, format="espresso-in", properties=properties, **kwargs)


read_espresso_dict = {"ph": read_espresso_ph, "pw": read_espresso_pw}

write_espresso_dict = {"ph": write_espresso_ph, "pw": write_espresso_pw}



"""Prepration for runners"""
from __future__ import annotations

import os
from datetime import datetime, timezone
from pathlib import Path
from shutil import move, rmtree
from tempfile import mkdtemp

from monty.shutil import gzip_dir, remove

from quacc import SETTINGS
from quacc.utils.files import (
    copy_decompress_files,
    copy_decompress_files_from_dir,
    make_unique_dir,
)


def calc_setup(copy_files: list[str | Path] | None = None) -> tuple[Path, Path]:
    """
    Perform staging operations for a calculation, including copying files to the scratch
    directory, setting the calculator's directory, decompressing files, and creating a
    symlink to the scratch directory.

    Parameters
    ----------
    copy_files
        Filenames to copy from source to scratch directory.

    Returns
    -------
    Path
        The path to the tmpdir, where the calculation will be run. It will be
        deleted after the calculation is complete. By default, this will be
        located within the `SETTINGS.SCRATCH_DIR`, but if that is not set, it will
        be located within the `SETTINGS.RESULTS_DIR`.
    Path
        The path to the results_dir, where the files will ultimately be stored.
        A symlink to the tmpdir will be made here during the calculation for
        convenience. By defualt, this will be the `SETTINGS.RESULTS_DIR`, but if
        `SETTINGS.CREATE_UNIQUE_DIR` is set, it will be a unique directory
        within the `SETTINGS.RESULTS_DIR`.
    """

    # Set where to store the results
    job_results_dir = (
        make_unique_dir(base_path=SETTINGS.RESULTS_DIR)
        if SETTINGS.CREATE_UNIQUE_DIR
        else SETTINGS.RESULTS_DIR
    )
    tmpdir_base = SETTINGS.SCRATCH_DIR or SETTINGS.RESULTS_DIR

    # Create a tmpdir for the calculation
    time_now = datetime.now(timezone.utc).strftime("%Y-%m-%d-%H-%M-%S-%f")
    tmpdir = Path(mkdtemp(prefix=f"quacc-tmp-{time_now}-", dir=tmpdir_base)).resolve()

    # Create a symlink to the tmpdir in the results_dir
    if os.name != "nt" and SETTINGS.SCRATCH_DIR:
        symlink = job_results_dir / f"{tmpdir.name}-symlink"
        symlink.unlink(missing_ok=True)
        symlink.symlink_to(tmpdir, target_is_directory=True)

    # Copy files to tmpdir and decompress them if needed
    if isinstance(copy_files, list):
        copy_decompress_files(copy_files, tmpdir)
    elif isinstance(copy_files, (str, Path)):
        if Path(copy_files).is_dir():
            copy_decompress_files_from_dir(copy_files, tmpdir)
        else:
            copy_decompress_files([copy_files], tmpdir)

    os.chdir(tmpdir)

    return tmpdir, job_results_dir


def calc_cleanup(tmpdir: Path, job_results_dir: Path) -> None:
    """
    Perform cleanup operations for a calculation, including gzipping files, copying
    files back to the original directory, and removing the tmpdir.

    Parameters
    ----------
    tmpdir
        The path to the tmpdir, where the calculation will be run. It will be
        deleted after the calculation is complete.
    job_results_dir
        The path to the job_results_dir, where the files will ultimately be
        stored. A symlink to the tmpdir will be made here during the calculation
        for convenience.

    Returns
    -------
    None
    """

    # Change to the results directory
    os.chdir(job_results_dir)

    # Gzip files in tmpdir
    if SETTINGS.GZIP_FILES:
        gzip_dir(tmpdir)

    # Move files from tmpdir to job_results_dir
    for file_name in os.listdir(tmpdir):
        if Path(job_results_dir / file_name).exists():
            remove(job_results_dir / file_name)
        move(tmpdir / file_name, job_results_dir / file_name)

    # Remove symlink to tmpdir
    symlink_path = job_results_dir / f"{tmpdir.name}-symlink"
    symlink_path.unlink(missing_ok=True)

    # Remove the tmpdir
    rmtree(tmpdir, ignore_errors=True)



"""Core recipes for ORCA."""
from __future__ import annotations

from typing import TYPE_CHECKING

import psutil

from quacc import job
from quacc.recipes.orca._base import base_fn

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Any

    from ase.atoms import Atoms

    from quacc.schemas._aliases.cclib import cclibSchema


@job
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    orcasimpleinput: dict[str, Any] | None = None,
    orcablocks: dict[str, Any] | None = None,
    nprocs: int | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> cclibSchema:
    """
    Carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    orcasimpleinput
        Dictionary of `orcasimpleinput` swaps for the calculator. To enable new
        entries, set the value as True. To remove entries from the defaults, set
        the value as None. For a list of available keys, refer to the
        `ase.calculators.orca.ORCA` calculator.
    orcablocks
        Dictionary of `orcablocks` swaps for the calculator. To enable new entries,
        set the value as True. To remove entries from the defaults, set the
        value as None. For a list of available keys, refer to the
        `ase.calculators.orca.ORCA` calculator.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    cclibSchema
        Dictionary of results from [quacc.schemas.cclib.cclib_summarize_run][]
    """

    nprocs = nprocs or psutil.cpu_count(logical=False)
    default_inputs = {
        xc: True,
        basis: True,
        "sp": True,
        "slowconv": True,
        "normalprint": True,
        "xyzfile": True,
    }
    default_blocks = {f"%pal nprocs {nprocs} end": True}

    return base_fn(
        atoms,
        charge,
        spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        additional_fields={"name": "ORCA Static"},
        copy_files=copy_files,
    )


@job
def relax_job(
    atoms: Atoms,
    charge: int,
    spin_multiplicity: int,
    xc: str = "wb97x-d3bj",
    basis: str = "def2-tzvp",
    run_freq: bool = False,
    orcasimpleinput: dict[str, Any] | None = None,
    orcablocks: dict[str, Any] | None = None,
    nprocs: int | None = None,
    copy_files: str | Path | list[str | Path] | None = None,
) -> cclibSchema:
    """
    Carry out a geometry optimization.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    xc
        Exchange-correlation functional
    basis
        Basis set
    run_freq
        If a frequency calculation should be carried out.
    orcasimpleinput
        Dictionary of `orcasimpleinput` swaps for the calculator. To enable new
        entries, set the value as True. To remove entries from the defaults, set
        the value as None. For a list of available keys, refer to the
        `ase.calculators.orca.ORCA` calculator.
    orcablocks
        Dictionary of `orcablocks` swaps for the calculator. To enable new entries,
        set the value as True. To remove entries from the defaults, set the
        value as None. For a list of available keys, refer to the
        `ase.calculators.orca.ORCA` calculator.
    nprocs
        Number of processors to use. Defaults to the number of physical cores.
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.

    Returns
    -------
    cclibSchema
        Dictionary of results from [quacc.schemas.cclib.cclib_summarize_run][]
    """

    nprocs = nprocs or psutil.cpu_count(logical=False)
    default_inputs = {
        xc: True,
        basis: True,
        "opt": True,
        "slowconv": True,
        "normalprint": True,
        "freq": True if run_freq else None,
        "xyzfile": True,
    }
    default_blocks = {f"%pal nprocs {nprocs} end": True}

    return base_fn(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        default_inputs=default_inputs,
        default_blocks=default_blocks,
        input_swaps=orcasimpleinput,
        block_swaps=orcablocks,
        additional_fields={"name": "ORCA Relax"},
        copy_files=copy_files,
    )



"""Utility functions for dealing with dictionaries."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def merge_dicts(
    dict1: dict[str, Any] | None,
    dict2: dict[str, Any] | None,
    remove_nones: bool = True,
) -> dict[str, Any]:
    """
    Recursively merges two dictionaries. If one of the inputs is `None`, then it is
    treated as `{}`.

    This function should be used instead of the | operator when merging nested dictionaries,
    e.g. `{"a": {"b": 1}} | {"a": {"c": 2}}` will return `{"a": {"c": 2}}` whereas
    `merge_dicts({"a": {"b": 1}}, {"a": {"c": 2}})` will return `{"a": {"b": 1, "c": 2}}`.

    Parameters
    ----------
    dict1
        First dictionary
    dict2
        Second dictionary
    remove_nones
        If True, remove empty lists and dictionaries

    Returns
    -------
    dict
        Merged dictionary
    """
    dict1 = dict1 or {}
    dict2 = dict2 or {}
    merged = dict1.copy()

    for key, value in dict2.items():
        if key in merged:
            if isinstance(merged[key], dict) and isinstance(value, dict):
                merged[key] = merge_dicts(merged[key], value)
            else:
                merged[key] = value
        else:
            merged[key] = value

    if remove_nones:
        merged = remove_dict_nones(merged)

    return merged


def merge_several_dicts(*args, remove_nones: bool = True) -> dict[str, Any]:
    """
    Recursively merge several dictionaries, taking the latter in the list as higher preference.

    Parameters
    ----------
    *args
        Dictionaries to merge
    remove_nones
        If True, remove empty lists and dictionaries

    Returns
    -------
    dict
        Merged dictionary
    """
    old_dict = args[0]
    for i in range(len(args) - 1):
        merged = merge_dicts(old_dict, args[i + 1], remove_nones=remove_nones)
        old_dict = merged.copy()
    return merged


def remove_dict_nones(start_dict: dict[str, Any]) -> dict[str, Any]:
    """
    For a given dictionary, recursively remove all items that are None.

    Parameters
    ----------
    start_dict
        Dictionary to clean

    Returns
    -------
    dict
        Cleaned dictionary
    """

    if isinstance(start_dict, dict):
        return {k: remove_dict_nones(v) for k, v in start_dict.items() if v is not None}
    return (
        [remove_dict_nones(v) for v in start_dict]
        if isinstance(start_dict, list)
        else start_dict
    )


def sort_dict(start_dict: dict[str, Any]) -> dict[str, Any]:
    """
    For a given dictionary, recursively sort all entries alphabetically by key.

    Parameters
    ----------
    start_dict
        Dictionary to sort

    Returns
    -------
    dict
        Sorted dictionary
    """

    return {
        k: sort_dict(v) if isinstance(v, dict) else v
        for k, v in sorted(start_dict.items())
    }



"""Core recipes for Psi4."""
from __future__ import annotations

from typing import TYPE_CHECKING

from monty.dev import requires

from quacc import job
from quacc.recipes.psi4._base import base_fn

try:
    import psi4
except ImportError:
    psi4 = None

if TYPE_CHECKING:
    from pathlib import Path

    from ase.atoms import Atoms

    from quacc.schemas._aliases.ase import RunSchema


@job
@requires(psi4, "Psi4 not installed. Try conda install -c psi4 psi4")
def static_job(
    atoms: Atoms,
    charge: int = 0,
    spin_multiplicity: int = 1,
    method: str = "wb97x-v",
    basis: str = "def2-tzvp",
    copy_files: str | Path | list[str | Path] | None = None,
    **kwargs,
) -> RunSchema:
    """
    Function to carry out a single-point calculation.

    Parameters
    ----------
    atoms
        Atoms object
    charge
        Charge of the system.
    spin_multiplicity
        Multiplicity of the system.
    method
        The level of theory to use.
    basis
        Basis set
    copy_files
        File(s) to copy to the runtime directory. If a directory is provided, it will be recursively unpacked.
    **kwargs
        Custom kwargs for the Psi4 calculator. Set a value to
        `None` to remove a pre-existing key entirely. For a list of available
        keys, refer to the `ase.calculators.psi4.Psi4` calculator.

    Returns
    -------
    RunSchema
        Dictionary of results from [quacc.schemas.ase.summarize_run][]
    """

    calc_defaults = {
        "mem": "16GB",
        "num_threads": "max",
        "method": method,
        "basis": basis,
        "charge": charge,
        "multiplicity": spin_multiplicity,
        "reference": "uks" if spin_multiplicity > 1 else "rks",
    }
    return base_fn(
        atoms,
        charge=charge,
        spin_multiplicity=spin_multiplicity,
        calc_defaults=calc_defaults,
        calc_swaps=kwargs,
        additional_fields={"name": "Psi4 Static"},
        copy_files=copy_files,
    )



"""Custodian handlers for VASP."""
from __future__ import annotations

import os
import shlex
from typing import TYPE_CHECKING

from custodian import Custodian
from custodian.vasp.handlers import (
    FrozenJobErrorHandler,
    IncorrectSmearingHandler,
    LargeSigmaHandler,
    MeshSymmetryErrorHandler,
    NonConvergingErrorHandler,
    PositiveEnergyErrorHandler,
    PotimErrorHandler,
    ScanMetalHandler,
    StdErrHandler,
    UnconvergedErrorHandler,
    VaspErrorHandler,
    WalltimeHandler,
)
from custodian.vasp.jobs import VaspJob
from custodian.vasp.validators import VaspFilesValidator, VasprunXMLValidator

if TYPE_CHECKING:
    from typing import Callable, TypedDict

    class VaspJobKwargs(TypedDict, total=False):
        output_file: str  # default = "vasp.out"
        stderr_file: str  # default = "std_err.txt"
        suffix: str  # default = ""
        final: bool  # default = True
        backup: bool  # default = True
        auto_npar: bool  # default = False
        auto_gamma: bool  # default = True
        settings_override: dict | None  # default = None
        copy_magmom: bool  # default = False
        auto_continue: bool  # default = False

    class CustodianKwargs(TypedDict, total=False):
        max_errors_per_job: int | None  # default = None
        polling_time_step: int  # default = 10
        monitor_freq: int  # default = 10
        skip_over_errors: bool  # default = False
        gzipped_output: bool  # default = False
        checkpoint: bool  # default = False
        terminate_func: Callable | None  # default = None
        terminate_on_nonzero_returncode: bool  # default = False


def run_custodian(
    vasp_parallel_cmd: str | None = None,
    vasp_cmd: str | None = None,
    vasp_gamma_cmd: str | None = None,
    vasp_custodian_max_errors: int | None = None,
    vasp_custodian_wall_time: float | None = None,
    vtst_fixes: bool | None = None,
    vasp_custodian_handlers: list[str] | None = None,
    vasp_custodian_validators: list[str] | None = None,
    scratch_dir: str | None = None,
    vasp_job_kwargs: VaspJobKwargs | None = None,
    custodian_kwargs: CustodianKwargs | None = None,
) -> None:
    """
    Function to run VASP Custodian.

    Parameters
    ----------
    vasp_parallel_cmd
        VASP parallel command, e.g. "srun -N 2 --ntasks-per-node=24". Defaults
        to the $VASP_PARALLEL_CMD environment variable in settings.
    vasp_cmd
        VASP command. Defaults to "vasp_std" in settings.
    vasp_gamma_cmd
        VASP gamma command. Defaults to "vasp_gam" in settings.
    vasp_custodian_max_errors
        Maximum number of errors to allow before stopping the run. Defaults to 5
        in settings.
    vasp_custodian_wall_time
        Maximum wall time to allow before creating a STOPCAR. Defaults to
        infinity in settings.
    vtst_fixes
        Whether to apply VTST input swaps. Defaults to False in settings.
    vasp_custodian_handlers
        List of handlers to use in Custodian. See settings for list.
    vasp_custodian_validators
        List of validators to use in Custodian. See settings for list.
    scratch_dir
        Scratch directory to use. Defaults to None.
    vasp_job_kwargs
        Keyword arguments to pass to the Custodian VaspJob. Defaults to None.
    custodian_kwargs
        Any remaining keyword arguments to pass to Custodian. Defaults to None.

    Returns
    -------
    None
    """
    # Adapted from atomate2.vasp.run.run_vasp

    from quacc import SETTINGS

    # Set defaults
    vasp_parallel_cmd = os.path.expandvars(
        SETTINGS.VASP_PARALLEL_CMD if vasp_parallel_cmd is None else vasp_parallel_cmd
    )
    vasp_cmd = SETTINGS.VASP_CMD if vasp_cmd is None else vasp_cmd
    vasp_gamma_cmd = (
        SETTINGS.VASP_GAMMA_CMD if vasp_gamma_cmd is None else vasp_gamma_cmd
    )
    vasp_custodian_max_errors = (
        SETTINGS.VASP_CUSTODIAN_MAX_ERRORS
        if vasp_custodian_max_errors is None
        else vasp_custodian_max_errors
    )
    vasp_custodian_wall_time = (
        SETTINGS.VASP_CUSTODIAN_WALL_TIME
        if vasp_custodian_wall_time is None
        else vasp_custodian_wall_time
    )
    vtst_fixes = SETTINGS.VASP_CUSTODIAN_VTST if vtst_fixes is None else vtst_fixes
    vasp_custodian_handlers = (
        SETTINGS.VASP_CUSTODIAN_HANDLERS
        if vasp_custodian_handlers is None
        else vasp_custodian_handlers
    )
    vasp_custodian_validators = (
        SETTINGS.VASP_CUSTODIAN_VALIDATORS
        if vasp_custodian_validators is None
        else vasp_custodian_validators
    )

    # Handlers for VASP
    handlers = []
    handlers_dict = {
        "VaspErrorHandler": VaspErrorHandler(vtst_fixes=vtst_fixes),
        "FrozenJobErrorHandler": FrozenJobErrorHandler(),
        "IncorrectSmearingHandler": IncorrectSmearingHandler(),
        "LargeSigmaHandler": LargeSigmaHandler(),
        "MeshSymmetryErrorHandler": MeshSymmetryErrorHandler(),
        "NonConvergingErrorHandler": NonConvergingErrorHandler(),
        "PositiveEnergyErrorHandler": PositiveEnergyErrorHandler(),
        "PotimErrorHandler": PotimErrorHandler(),
        "StdErrHandler": StdErrHandler(),
        "UnconvergedErrorHandler": UnconvergedErrorHandler(),
        "WalltimeHandler": WalltimeHandler(),
        "ScanMetalHandler": ScanMetalHandler(),
    }
    validators_dict = {
        "VaspFilesValidator": VaspFilesValidator(),
        "VasprunXMLValidator": VasprunXMLValidator(),
    }

    handlers = []
    for handler_flag in vasp_custodian_handlers:
        if handler_flag not in handlers_dict:
            msg = f"Unknown VASP error handler: {handler_flag}"
            raise ValueError(msg)
        handlers.append(handlers_dict[handler_flag])

    validators = []
    for validator_flag in vasp_custodian_validators:
        if validator_flag not in validators_dict:
            msg = f"Unknown VASP validator: {validator_flag}"
            raise ValueError(msg)
        validators.append(validators_dict[validator_flag])

    # Populate settings
    full_vasp_cmd = f"{vasp_parallel_cmd} {vasp_cmd}"
    full_vasp_gamma_cmd = f"{vasp_parallel_cmd} {vasp_gamma_cmd}"

    # Run VASP
    vasp_job_kwargs = {} if vasp_job_kwargs is None else vasp_job_kwargs
    custodian_kwargs = {} if custodian_kwargs is None else custodian_kwargs
    split_vasp_cmd = shlex.split(full_vasp_cmd)
    split_vasp_gamma_cmd = shlex.split(full_vasp_gamma_cmd)
    vasp_job_kwargs["gamma_vasp_cmd"] = split_vasp_gamma_cmd

    # Run with Custodian
    jobs = [VaspJob(split_vasp_cmd, **vasp_job_kwargs)]

    if vasp_custodian_wall_time:
        handlers = [
            *list(handlers),
            WalltimeHandler(wall_time=vasp_custodian_wall_time),
        ]

    c = Custodian(
        handlers,
        jobs,
        validators=validators,
        max_errors=vasp_custodian_max_errors,
        scratch_dir=scratch_dir,
        **custodian_kwargs,
    )

    c.run()


if __name__ == "__main__":
    run_custodian()
